"""Ensemble statistics and weights."""

import inspect
import logging
import os
import warnings
from copy import deepcopy
from itertools import chain, groupby
from pathlib import Path

import numpy as np
import xarray as xr
from xclim import ensembles

from .catalog import DataCatalog
from .catutils import generate_id
from .config import parse_config
from .indicators import compute_indicators
from .regrid import regrid_dataset
from .spatial import get_grid_mapping, subset
from .utils import clean_up, get_cat_attrs

logger = logging.getLogger(__name__)

__all__ = [
    "build_partition_data",
    "ensemble_stats",
    "generate_weights",
    "reduce_ensemble",
]


@parse_config
def ensemble_stats(  # noqa: C901
    datasets: (
        dict
        | list[str | os.PathLike]
        | list[xr.Dataset]
        | list[xr.DataArray]
        | xr.Dataset
    ),
    statistics: dict,
    *,
    create_kwargs: dict | None = None,
    weights: xr.DataArray | None = None,
    common_attrs_only: bool = True,
    to_level: str = "ensemble",
) -> xr.Dataset:
    """Create an ensemble and computes statistics on it.

    Parameters
    ----------
    datasets : dict or list of [str, os.PathLike, Dataset or DataArray], or Dataset
        List of file paths or xarray Dataset/DataArray objects to include in the ensemble.
        A dictionary can be passed instead of a list, in which case the keys are used as coordinates along the new
        `realization` axis.
        Tip: With a project catalog, you can do: `datasets = pcat.search(**search_dict).to_dataset_dict()`.
        If a single Dataset is passed, it is assumed to already be an ensemble and will be used as is. The 'realization' dimension is required.
    statistics : dict
        xclim.ensembles statistics to be called. Dictionary in the format {function: arguments}.
        If a function requires 'weights', you can leave it out of this dictionary and
        it will be applied automatically if the 'weights' argument is provided.
        See the Notes section for more details on robustness statistics, which are more complex in their usage.
    create_kwargs : dict, optional
        Dictionary of arguments for xclim.ensembles.create_ensemble.
    weights : xr.DataArray, optional
        Weights to apply along the 'realization' dimension. This array cannot contain missing values.
    common_attrs_only : bool
        If True, keeps only the global attributes that are the same for all datasets and generate new id.
        If False, keeps global attrs of the first dataset (same behaviour as xclim.ensembles.create_ensemble)
    to_level : str
        The processing level to assign to the output.

    Returns
    -------
    xr.Dataset
        Dataset with ensemble statistics

    Notes
    -----
    * The positive fraction in 'change_significance' and 'robustness_fractions' is calculated by
      xclim using 'v > 0', which is not appropriate for relative deltas.
      This function will attempt to detect relative deltas by using the 'delta_kind' attribute ('rel.', 'relative', '*', or '/')
      and will apply 'v - 1' before calling the function.
    * The 'robustness_categories' statistic requires the outputs of 'robustness_fractions'.
      Thus, there are two ways to build the 'statistics' dictionary:

      1. Having 'robustness_fractions' and 'robustness_categories' as separate entries in the dictionary.
         In this case, all outputs will be returned.
      2. Having 'robustness_fractions' as a nested dictionary under 'robustness_categories'.
         In this case, only the robustness categories will be returned.

    * A 'ref' DataArray can be passed to 'change_significance' and 'robustness_fractions', which will be used by xclim to compute deltas
      and perform some significance tests. However, this supposes that both 'datasets' and 'ref' are still timeseries (e.g. annual means),
      not climatologies where the 'time' dimension represents the period over which the climatology was computed. Thus,
      using 'ref' is only accepted if 'robustness_fractions' (or 'robustness_categories') is the only statistic being computed.
    * If you want to use compute a robustness statistic on a climatology, you should first compute the climatologies and deltas yourself,
      then leave 'ref' as None and pass the deltas as the 'datasets' argument. This will be compatible with other statistics.

    See Also
    --------
    xclim.ensembles._base.create_ensemble, xclim.ensembles._base.ensemble_percentiles,
    xclim.ensembles._base.ensemble_mean_std_max_min,
    xclim.ensembles._robustness.robustness_fractions, xclim.ensembles._robustness.robustness_categories,
    xclim.ensembles._robustness.robustness_coefficient,
    """
    create_kwargs = create_kwargs or {}
    statistics = deepcopy(statistics)  # to avoid modifying the original dictionary

    # if input files are .zarr, change the engine automatically
    if isinstance(datasets, list) and isinstance(datasets[0], str | os.PathLike):
        path = Path(datasets[0])
        if path.suffix in [".zarr", ".zip"]:
            create_kwargs.setdefault("engine", "zarr")

    if not isinstance(datasets, xr.Dataset):
        ens = ensembles.create_ensemble(datasets, **create_kwargs)
    else:
        ens = datasets

    ens_stats = xr.Dataset(attrs=ens.attrs)

    # "robustness_categories" requires "robustness_fractions", but we want to compute things only once if both are requested.
    statistics_to_compute = list(statistics.keys())
    if "robustness_categories" in statistics_to_compute:
        if "robustness_fractions" in statistics_to_compute:
            statistics_to_compute.remove("robustness_fractions")
        elif "robustness_fractions" not in statistics["robustness_categories"]:
            raise ValueError(
                "'robustness_categories' requires 'robustness_fractions' to be computed. "
                "Either add 'robustness_fractions' to the statistics dictionary or "
                "add 'robustness_fractions' under the 'robustness_categories' dictionary."
            )

    for stat in statistics_to_compute:
        stats_kwargs = deepcopy(statistics.get(stat) or {})
        msg = f"Calculating {stat} from an ensemble of {len(ens.realization)} simulations."
        logger.info(msg)

        # Workaround for robustness_categories
        real_stat = None
        categories_kwargs = {}
        if stat == "robustness_categories":
            real_stat = "robustness_categories"
            stat = "robustness_fractions"
            categories_kwargs = deepcopy(stats_kwargs)
            categories_kwargs.pop("robustness_fractions", None)
            stats_kwargs = deepcopy(
                stats_kwargs.get("robustness_fractions", None)
                or statistics.get("robustness_fractions", {})
            )

        if weights is not None:
            if "weights" in inspect.getfullargspec(getattr(ensembles, stat))[0]:
                stats_kwargs["weights"] = weights.reindex_like(ens.realization)
            else:
                warnings.warn(
                    f"Weighting is not supported for '{stat}'. The results may be incorrect."
                )

        if stat in [
            "robustness_fractions",
            "robustness_categories",
        ]:
            # These statistics only work on DataArrays
            for v in ens.data_vars:
                with xr.set_options(keep_attrs=True):
                    # Support for relative deltas [0, inf], where positive fraction is 'v > 1' instead of 'v > 0'.
                    delta_kind = ens[v].attrs.get("delta_kind")
                    if stats_kwargs.get("ref") is not None and delta_kind is not None:
                        raise ValueError(
                            f"{v} is a delta, but 'ref' was still specified."
                        )
                    if delta_kind in ["rel.", "relative", "*", "/"]:
                        msg = f"Relative delta detected for {v}. Applying 'v - 1' before change_significance."
                        logging.info(msg)
                        ens_v = ens[v] - 1
                    else:
                        ens_v = ens[v]

                    # Call the function
                    tmp = getattr(ensembles, stat)(ens_v, **stats_kwargs)

                    # Robustness categories
                    if real_stat == "robustness_categories":
                        categories = ensembles.robustness_categories(
                            tmp, **categories_kwargs
                        )
                        ens_stats[f"{v}_robustness_categories"] = categories

                    # Only return the robustness fractions if they were requested.
                    if "robustness_fractions" in statistics.keys():
                        tmp = tmp.rename({s: f"{v}_{s}" for s in tmp.data_vars})
                        ens_stats = ens_stats.merge(tmp)

        else:
            ens_stats = ens_stats.merge(getattr(ensembles, stat)(ens, **stats_kwargs))

    # delete the realization coordinate if there
    if "realization" in ens_stats:
        ens_stats = ens_stats.drop_vars("realization")

    # delete attrs that are not common to all dataset
    if common_attrs_only and not isinstance(datasets, xr.Dataset):
        # if they exist remove attrs specific to create_ensemble
        create_kwargs.pop("mf_flag", None)
        create_kwargs.pop("resample_freq", None)
        create_kwargs.pop("calendar", None)
        create_kwargs.pop("preprocess", None)

        ens_stats = clean_up(
            ds=ens_stats,
            common_attrs_only=datasets,
            common_attrs_open_kwargs=create_kwargs,
        )

    ens_stats.attrs["cat:processing_level"] = to_level
    ens_stats.attrs["ensemble_size"] = len(datasets)

    return ens_stats


def generate_weights(  # noqa: C901
    datasets: dict | list,
    *,
    independence_level: str = "model",
    balance_experiments: bool = False,
    attribute_weights: dict | None = None,
    skipna: bool = True,
    v_for_skipna: str | None = None,
    standardize: bool = False,
) -> xr.DataArray:
    """Use realization attributes to automatically generate weights along the 'realization' dimension.

    Parameters
    ----------
    datasets : dict
        List of Dataset objects that will be included in the ensemble.
        The datasets should include the necessary attributes to understand their metadata - See 'Notes' below.
        A dictionary can be passed instead of a list, in which case the keys are used for the 'realization' coordinate.
        Tip: With a project catalog, you can do: `datasets = pcat.search(**search_dict).to_dataset_dict()`.
    independence_level : str
        'model': Weights using the method '1 model - 1 Vote',
        where every unique combination of 'source' and 'driving_model' is considered a model.
        'GCM': Weights using the method '1 GCM - 1 Vote'
        'institution': Weights using the method '1 institution - 1 Vote'
    balance_experiments : bool
        If True, each experiment will be given a total weight of 1
        (prior to subsequent weighting made through `attribute_weights`).
        This option requires the 'cat:experiment' attribute to be present in all datasets.
    attribute_weights : dict, optional
        Nested dictionaries of weights to apply to each dataset.
        These weights are applied after the independence weighting.
        The first level of keys are the attributes for which weights are being given.
        The second level of keys are unique entries for the attribute, with the value being either an individual weight
        or a xr.DataArray. If a DataArray is used, its dimensions must be the same non-stationary coordinate
        as the datasets (ex: time, horizon) and the attribute being weighted (ex: experiment).
        A `others` key can be used to give the same weight to all entries not specifically named in the dictionary.
        Example #1: {'source': {'MPI-ESM-1-2-HAM': 0.25, 'MPI-ESM1-2-HR': 0.5}},
        Example #2: {'experiment': {'ssp585': xr.DataArray, 'ssp126': xr.DataArray}, 'institution': {'CCCma': 0.5, 'others': 1}}
    skipna : bool
        If True, weights will be computed from attributes only.
        If False, weights will be computed from the number of non-missing values.
        skipna=False requires either a 'time' or 'horizon' dimension in the datasets.
    v_for_skipna : str, optional
        Variable to use for skipna=False. If None, the first variable in the first dataset is used.
    standardize : bool
        If True, the weights are standardized to sum to 1 (per timestep/horizon, if skipna=False).

    Notes
    -----
    The following attributes are required for the function to work:
        - 'cat:source' in all datasets
        - 'cat:driving_model' in regional climate models
        - 'cat:institution' in all datasets if independence_level='institution'
        - 'cat:experiment' in all datasets if split_experiments=True

    Even when not required, the 'cat:member' and 'cat:experiment' attributes are strongly recommended to ensure the weights are computed correctly.

    Returns
    -------
    xr.DataArray
        Weights along the 'realization' dimension, or 2D weights along the 'realization' and 'time/horizon' dimensions if skipna=False.
    """
    if isinstance(datasets, list):
        datasets = {i: datasets[i] for i in range(len(datasets))}

    if independence_level not in ["model", "GCM", "institution"]:
        raise ValueError(
            f"'independence_level' should be between 'model', 'GCM', and 'institution', received {independence_level}."
        )
    if skipna is False:
        if v_for_skipna is None:
            v_for_skipna = list(datasets[list(datasets.keys())[0]].data_vars)[0]
            msg = f"Using '{v_for_skipna}' as the variable to check for missing values."
            logger.info(msg)

        # Check if any dataset has dimensions that are not 'time' or 'horizon'
        other_dims = {
            k: [
                d
                for d in datasets[k][v_for_skipna].dims
                if d not in ["time", "horizon"]
            ]
            for k in datasets.keys()
        }
        for k in other_dims:
            if len(other_dims[k]) > 0:
                warnings.warn(
                    f"Dataset {k} has dimensions that are not 'time' or 'horizon': {other_dims[k]}. "
                    "The first indexes of these dimensions will be used to compute the weights."
                )
                datasets[k] = datasets[k].isel({d: 0 for d in other_dims[k]})

    # Use metadata to identify the simulation attributes
    keys = datasets.keys()
    defdict = {
        "experiment": None,
        "institution": None,
        "driving_model": None,
        "source": None,
        "member": None,
    }

    info = {key: dict(defdict, **get_cat_attrs(datasets[key])) for key in keys}

    # Check if there are both RCMs and GCMs in datasets, with attribute_weights set to weight them.
    if attribute_weights and (
        any(a in ["source", "driving_model"] for a in list(attribute_weights.keys()))
        and len(list(groupby([info[k]["driving_model"] is None for k in info.keys()])))
        > 1
    ):
        raise NotImplementedError(
            "Weighting `source` and/or `driving_model` through `attribute_weights` "
            "is not yet implemented when given a mix of GCMs and RCMs."
        )

    # More easily manage GCMs and RCMs
    for k in info:
        if info[k]["driving_model"] is None or len(info[k]["driving_model"]) == 0:
            info[k]["driving_model"] = info[k]["source"]

    # Verifications
    if any(
        (info[k]["driving_model"] is None or len(info[k]["driving_model"]) == 0)
        for k in info
    ):
        raise ValueError(
            "The 'cat:source' or 'cat:driving_model' attribute is missing from some simulations."
        )
    if balance_experiments and any(
        (info[k]["experiment"] is None or len(info[k]["experiment"]) == 0) for k in info
    ):
        raise ValueError(
            "The 'cat:experiment' attribute is missing from some simulations. 'balance_experiments' cannot be True."
        )
    if independence_level == "institution" and any(
        (info[k]["institution"] is None or len(info[k]["institution"]) == 0)
        for k in info
    ):
        raise ValueError(
            "The 'cat:institution' attribute is missing from some simulations. 'independence_level' cannot be 'institution'."
        )
    for attr in ["member", "experiment"]:
        if any(info[k][attr] is None for k in info):
            if all(info[k][attr] is None for k in info):
                warnings.warn(
                    f"The 'cat:{attr}' attribute is missing from all datasets. Make sure the results are correct."
                )
            else:
                warnings.warn(
                    f"The 'cat:{attr}' attribute is inconsistent across datasets. Results are likely to be incorrect."
                )

    # Combine the member and experiment attributes
    for k in info:
        info[k]["member-exp"] = (
            str(info[k]["member"]) + "-" + str(info[k]["experiment"])
        )

    # Build the weights according to the independence structure
    if skipna:
        weights = xr.DataArray(
            np.zeros(len(info.keys())),
            dims=["realization"],
            coords={"realization": list(info.keys())},
        )
    else:
        # Get the name of the extra dimension
        extra_dim = list(
            chain.from_iterable(
                [
                    [
                        datasets[list(keys)[d]][h]
                        for h in ["time", "horizon"]
                        if h in datasets[list(keys)[d]].dims
                    ]
                    for d in range(len(keys))
                ]
            )
        )
        if len({e.name for e in extra_dim}) != 1:
            raise ValueError(
                f"Expected either 'time' or 'horizon' as an extra dimension, found {extra_dim}."
            )

        # Combine the extra dimension and remove duplicates
        extra_dimension = xr.concat(extra_dim, dim=extra_dim[0].name).drop_duplicates(
            extra_dim[0].name
        )

        # Check that the extra dimension is the same for all datasets.
        # If not, modify the datasets to make them the same.
        if not all(extra_dimension.equals(extra_dim[d]) for d in range(len(extra_dim))):
            warnings.warn(
                f"Extra dimension {extra_dimension.name} is not the same for all datasets. Reindexing."
            )
            for d in datasets.keys():
                datasets[d] = datasets[d].reindex(
                    {extra_dimension.name: extra_dimension}
                )

        weights = xr.DataArray(
            np.zeros((len(info.keys()), len(extra_dimension))),
            dims=["realization", extra_dimension.name],
            coords={
                "realization": list(info.keys()),
                extra_dimension.name: extra_dimension,
            },
        )

    for i in range(len(info)):
        sim = info[list(keys)[i]]

        # Number of models running a given realization of a driving model
        models_struct = (
            ["source", "driving_model", "member-exp"]
            if independence_level == "model"
            else ["driving_model", "member-exp"]
        )
        models = [
            k for k in info.keys() if all([info[k][s] == sim[s] for s in models_struct])
        ]

        if skipna:
            n_models = len(models)
        else:
            n_models = xr.concat(
                [datasets[k][v_for_skipna].notnull() for k in models], dim="realization"
            ).sum(dim="realization")

        # Number of realizations of a given driving model
        if independence_level == "model":
            realization_struct = (
                ["source", "driving_model", "experiment"]
                if balance_experiments
                else ["source", "driving_model"]
            )
        else:
            realization_struct = (
                ["driving_model", "experiment"]
                if balance_experiments
                else ["driving_model"]
            )
        realizations = {
            info[k]["member-exp"]
            for k in info.keys()
            if all([info[k][s] == sim[s] for s in realization_struct])
        }

        if skipna:
            n_realizations = len(realizations)
        else:
            n_realizations = xr.zeros_like(datasets[list(keys)[0]][v_for_skipna])
            r_models = dict()
            for r in realizations:
                r_models[r] = [
                    k
                    for k in info.keys()
                    if (
                        all([info[k][s] == sim[s] for s in realization_struct])
                        and (info[k]["member-exp"] == r)
                    )
                ]
                n_realizations = n_realizations + (
                    xr.concat(
                        [datasets[k][v_for_skipna].notnull() for k in r_models[r]],
                        dim="realization",
                    ).sum(dim="realization")
                    > 0
                )

        # Number of driving models run by a given institution
        if independence_level == "institution":
            institution_struct = (
                ["institution", "experiment"]
                if balance_experiments
                else ["institution"]
            )
            institution = {
                info[k]["driving_model"]
                for k in info.keys()
                if all([info[k][s] == sim[s] for s in institution_struct])
            }

            if skipna:
                n_institutions = len(institution)
            else:
                n_institutions = xr.zeros_like(datasets[list(keys)[0]][v_for_skipna])
                i_models = dict()
                for ii in institution:
                    i_models[ii] = [
                        k
                        for k in info.keys()
                        if (
                            all([info[k][s] == sim[s] for s in institution_struct])
                            and (info[k]["driving_model"] == ii)
                        )
                    ]
                    n_institutions = n_institutions + (
                        xr.concat(
                            [datasets[k][v_for_skipna].notnull() for k in i_models[ii]],
                            dim="realization",
                        ).sum(dim="realization")
                        > 0
                    )
        else:
            n_institutions = 1

        # Divide the weight equally between the group
        w = 1 / n_models / n_realizations / n_institutions
        weights[i] = xr.where(np.isfinite(w), w, 0)

    if balance_experiments:
        # Divide the weight equally between the experiments
        experiments = [info[k]["experiment"] for k in info.keys()]
        weights = weights.assign_coords(
            experiment=("realization", experiments),
        )
        expsum = weights.groupby("experiment").sum(dim="realization")

        for e in expsum.experiment:
            weights = weights.where(
                weights.experiment != e, other=weights / expsum.sel(experiment=e)
            )

        # Drop the experiment coordinate
        weights = weights.drop_vars("experiment")

    # Attribute_weights
    if attribute_weights:
        stationary_weights = {}
        non_stationary_weights = {}
        for att, v_att in attribute_weights.items():
            # Add warning when a mismatch between independance_level/experiment_weight and attribute_weights is detected
            if att == "experiment" and not balance_experiments:
                warnings.warn(
                    "Key experiment given in attribute_weights without argument balance_experiments=True"
                )

            if (
                (att == "source" and independence_level != "model")
                or (att == "driving_model" and independence_level != "GCM")
                or (att == "institution" and independence_level != "institution")
            ):
                warnings.warn(
                    f"The {att} weights do not match the {independence_level} independence_level"
                )

            # Verification
            if att not in info[k] or any(
                (info[k][att] is None or len(info[k][att]) == 0) for k in info
            ):
                raise ValueError(
                    f"The {att} attribute is missing from some simulations."
                )
            # Split dict and xr.DataArray weights
            if isinstance(v_att, xr.DataArray):
                non_stationary_weights[att] = v_att
            elif isinstance(v_att, dict):
                stationary_weights[att] = v_att
            else:
                raise ValueError("Attribute_weights should be dict or xr.DataArray.")
        # Stationary weights (dicts)
        if stationary_weights:
            for att, w_dict in stationary_weights.items():
                for k, v in info.items():
                    if v[att] not in w_dict and "others" not in w_dict:
                        raise ValueError(
                            f"The {att} {v[att]} or others are not in the attribute_weights dict."
                        )
                    elif v[att] not in w_dict and "others" in w_dict:
                        w = w_dict["others"]
                    elif v[att] in w_dict:
                        w = w_dict[v[att]]
                    weights.loc[{"realization": k}] = weights.sel(realization=k) * w
        # Non-stationary weights (xr.DataArray)
        if non_stationary_weights:
            for att, da in non_stationary_weights.items():
                # check if the attribute is in the xr.DataArray coords
                if att not in da.coords:
                    raise ValueError(f"{att} is not in the xr.DataArray coords.")
                # find the coordinate (coord) to broadcast the weights (not equal to the attribute), ex: time / horizon
                ls_coord = list(da.coords)
                ls_coord.remove(att)
                if len(ls_coord) > 1:
                    raise ValueError(
                        f"The {att} DataArray has more than one coord dimension to apply weights."
                    )
                else:
                    coord = ls_coord[0]
                # broadcast coord to the weights DataArray
                if coord not in weights.coords:
                    weights = weights.expand_dims({coord: da[coord].values})
                ls_da = []
                for k, v in info.items():
                    if v[att] not in da[att] and "others" not in da[att]:
                        raise ValueError(
                            f"The {att} {v[att]} or others are not in the attribute_weights datarray coords."
                        )
                    elif v[att] not in da[att] and "others" in da[att]:
                        ls_da.append(da.sel(**{att: "others"}).drop_vars(att))
                    else:
                        ls_da.append(da.sel(**{att: v[att]}).drop_vars(att))
                nw = xr.concat(ls_da, weights.realization)
                weights = weights * nw

    if standardize:
        weights = weights / weights.sum(dim="realization")

    return weights


def _partition_from_list(datasets, partition_dim, subset_kw, regrid_kw):
    list_ds = []
    # only keep attrs common to all datasets
    common_attrs = False
    for ds in datasets:
        if subset_kw:
            ds = subset(ds, **subset_kw)
            gridmap = get_grid_mapping(ds)
            ds = ds.drop_vars(
                [
                    ds.cf["longitude"].name,
                    ds.cf["latitude"].name,
                    ds.cf.axes["X"][0],
                    ds.cf.axes["Y"][0],
                    gridmap,
                ],
                errors="ignore",
            )

        if regrid_kw:
            ds = regrid_dataset(ds, **regrid_kw)

        for dim in partition_dim:
            if f"cat:{dim}" in ds.attrs:
                ds = ds.expand_dims(**{dim: [ds.attrs[f"cat:{dim}"]]})

        if "bias_adjust_project" in ds.dims:
            ds = ds.assign_coords(
                adjustment=(
                    "bias_adjust_project",
                    [ds.attrs.get("cat:adjustment", np.nan)],
                )
            )
            ds = ds.assign_coords(
                reference=(
                    "bias_adjust_project",
                    [ds.attrs.get("cat:reference", np.nan)],
                )
            )

        if "realization" in partition_dim:
            new_source = f"{ds.attrs['cat:institution']}_{ds.attrs['cat:source']}_{ds.attrs['cat:member']}"
            ds = ds.expand_dims(realization=[new_source])

        a = ds.attrs
        a.pop("intake_esm_vars", None)  # remove list for intersection to work
        common_attrs = dict(common_attrs.items() & a.items()) if common_attrs else a
        list_ds.append(ds)
    ens = xr.merge(list_ds)
    ens.attrs = common_attrs
    return ens


def _partition_from_catalog(
    datasets, partition_dim, subset_kw, regrid_kw, to_dataset_kw
):

    if ("adjustment" in partition_dim or "reference" in partition_dim) and (
        "bias_adjust_project" in partition_dim
    ):
        raise ValueError(
            "The partition_dim can have either adjustment and reference or bias_adjust_project, not both."
        )

    if ("realization" in partition_dim) and ("source" in partition_dim):
        raise ValueError(
            "The partition_dim can have either realization or source, not both."
        )

    # special case to handle source (create one dimension with institution_source_member)
    ensemble_on_list = None
    if "realization" in partition_dim:
        partition_dim.remove("realization")
        ensemble_on_list = ["institution", "source", "member"]

    subcat = datasets

    # get attrs that are common to all datasets
    common_attrs = {}
    for col, series in subcat.df.items():
        if (series[0] == series).all():
            common_attrs[f"cat:{col}"] = series[0]

    col_id = [
        (
            "adjustment" if "adjustment" in partition_dim else None
        ),  # instead of bias_adjust_project, need to use adjustment, not method bc .sel
        (
            "reference" if "reference" in partition_dim else None
        ),  # instead of bias_adjust_project
        "bias_adjust_project" if "bias_adjust_project" in partition_dim else None,
        "mip_era",
        "activity",
        "driving_model",
        "institution" if "realization" in partition_dim else None,
        "source",
        "experiment",
        "member" if "realization" in partition_dim else None,
        "domain",
    ]

    subcat.df["id"] = generate_id(subcat.df, col_id)

    # create a dataset for each bias_adjust_project, modify grid and concat them
    # choose dim that exists in partition_dim and first in the order of preference
    order_of_preference = ["reference", "bias_adjust_project", "source"]
    dim_with_different_grid = list(set(partition_dim) & set(order_of_preference))[0]

    list_ds = []
    for d in subcat.df[dim_with_different_grid].unique():
        ds = subcat.search(**{dim_with_different_grid: d}).to_dataset(
            concat_on=partition_dim,
            create_ensemble_on=ensemble_on_list,
            **to_dataset_kw,
        )

        if subset_kw:
            ds = subset(ds, **subset_kw)
            gridmap = get_grid_mapping(ds)
            ds = ds.drop_vars(
                [
                    ds.cf["longitude"].name,
                    ds.cf["latitude"].name,
                    ds.cf.axes["X"][0],
                    ds.cf.axes["Y"][0],
                    gridmap,
                ],
                errors="ignore",
            )

        if regrid_kw:
            ds = regrid_dataset(ds, **regrid_kw)

        # add coords adjustment and reference
        if "bias_adjust_project" in ds.dims:
            ds = ds.assign_coords(
                adjustment=(
                    "bias_adjust_project",
                    [ds.attrs.get("cat:adjustment", np.nan)],
                )
            )  # need to use adjustment, not method bc .sel
            ds = ds.assign_coords(
                reference=(
                    "bias_adjust_project",
                    [ds.attrs.get("cat:reference", np.nan)],
                )
            )
        list_ds.append(ds)
    ens = xr.concat(list_ds, dim=dim_with_different_grid)
    ens.attrs = common_attrs
    return ens


def build_partition_data(
    datasets: dict | list[xr.Dataset],
    partition_dim: list[str] = ["realization", "experiment", "bias_adjust_project"],
    subset_kw: dict | None = None,
    regrid_kw: dict | None = None,
    rename_dict: dict | None = None,
    to_dataset_kw: dict | None = None,
    to_level: str = "partition-ensemble",
):
    """
    Get the input for the xclim partition functions.

    From a list or dictionary of datasets, create a single dataset with
    `partition_dim` dimensions (and time) to pass to one of the xclim partition functions
    (https://xclim.readthedocs.io/en/stable/api.html#uncertainty-partitioning).
    If the inputs have different grids,
    they have to be subsetted and/or regridded to a common grid/point.

    Parameters
    ----------
    datasets : list[xr.Dataset], dict[str, xr.Dataset], DataCatalog
        Either a list/dictionary of Datasets or a DataCatalog that will be included in the ensemble.
        The datasets should include the necessary ("cat:") attributes to understand their metadata.
        Tip: A dictionary can be created with `datasets = pcat.search(**search_dict).to_dataset_dict()`.

        The use of a DataCatalog is recommended for large ensembles.
        In that case, the ensembles will be loaded separately for each `bias_adjust_project`,
        the subsetting or regridding can be applied before combining the datasets through concatenation.
        If `bias_adjust_project` is not in `partition_dim`, `source` will be used instead.
    partition_dim: list[str]
        Components of the partition. They will become the dimension of the output.
        The default is ['source', 'experiment', 'bias_adjust_project'].
        For source, the dimension will actually be institution_source_member.
    subset_kw : dict, optional
        Arguments to pass to `xs.spatial.subset()`.
    regrid_kw : dict, optional
        Arguments to pass to `xs.regrid_dataset()`.
        Note that regriding is computationally expensive. For large datasets,
        it might be worth it to do the regridding first, outside of this function.
    rename_dict : dict, optional
        Dictionary to rename the dimensions from xscen names to xclim names.
        The default is {'source': 'model', 'bias_adjust_project': 'downscaling', 'experiment': 'scenario'}.
    to_dataset_kw : dict, optional
        Arguments to pass to `xscen.DataCatalog.to_dataset()` if datasets is a DataCatalog.
    to_level: str
        The processing level of the output dataset. Default is 'partition-ensemble'.

    Returns
    -------
    xr.Dataset
        The input data for the partition functions.

    See Also
    --------
    xclim.ensembles
    """
    if isinstance(datasets, dict):
        datasets = list(datasets.values())
    # initialize dict
    subset_kw = subset_kw or {}
    regrid_kw = regrid_kw or {}
    to_dataset_kw = to_dataset_kw or {}

    if isinstance(datasets, list):
        ens = _partition_from_list(datasets, partition_dim, subset_kw, regrid_kw)

    elif isinstance(datasets, DataCatalog):
        ens = _partition_from_catalog(
            datasets, partition_dim, subset_kw, regrid_kw, to_dataset_kw
        )

    else:
        raise ValueError(
            "'datasets' should be a list/dictionary of xarray datasets or a xscen.DataCatalog"
        )

    rename_dict = rename_dict or {}
    rename_dict.setdefault("realization", "model")
    rename_dict.setdefault("source", "model")
    rename_dict.setdefault("experiment", "scenario")
    rename_dict.setdefault("bias_adjust_project", "downscaling")
    rename_dict = {k: v for k, v in rename_dict.items() if k in ens.dims}
    ens = ens.rename(rename_dict)

    ens.attrs["cat:processing_level"] = to_level
    ens.attrs["cat:id"] = generate_id(ens)[0]

    return ens


@parse_config
def reduce_ensemble(
    data: xr.DataArray | dict | list | xr.Dataset,
    method: str,
    *,
    horizons: list[str] | None = None,
    create_kwargs: dict | None = None,
    **kwargs,
):
    r"""Reduce an ensemble of simulations using clustering algorithms from xclim.ensembles.

    Parameters
    ----------
    data : xr.DataArray
        Selection criteria data : 2-D xr.DataArray with dimensions 'realization' and 'criteria'.
        These are the values used for clustering. Realizations represent the individual original
        ensemble members and criteria the variables/indicators used in the grouping algorithm.
        This data can be generated using py:func:`xclim.ensembles.make_criteria`.
        Alternatively, either a xr.Dataset, a list of xr.Dataset or a dictionary of xr.Dataset can be passed,
        in which case the data will be built using py:func:`xclim.ensembles.create_ensemble` and py:func:`xclim.ensembles.make_criteria`.
    method : str
      ['kkz', 'kmeans']. Clustering method.
    horizons : list of str, optional
        Subset of horizons on which to create the data. Only used if `data` needs to be built.
    create_kwargs : dict, optional
        Arguments to pass to py:func:`xclim.ensembles.create_ensemble` if `data` is not an xr.DataArray.
    \*\*kwargs : dict
        Arguments to send to either py:func:`xclim.ensembles.kkz_reduce_ensemble` or py:func:`xclim.ensembles.kmeans_reduce_ensemble`.

    Returns
    -------
    selected : xr.DataArray
        DataArray of dimension 'realization' with the selected simulations.
    clusters : dict
        If using kmeans clustering, realizations grouped by cluster.
    fig_data : dict
        If using kmeans clustering, data necessary to call py:func:`xclim.ensembles.plot_rsqprofile`.

    Notes
    -----
    If building `data` to be constructed by this function, the datasets should already have a climatology computed on them, such that the data
    has no temporal dimension aside from the "horizon" coordinate (which is optional and might be used to subset the data).
    If the indicators are a mix of yearly, seasonal, and monthly, they should be stacked on the same time/horizon axis and put in the same dataset.
    You can use py:func:`xscen.utils.unstack_dates` on seasonal or monthly indicators to this end.
    """
    if isinstance(data, list | dict):
        data = ensembles.create_ensemble(datasets=data, **(create_kwargs or {}))
    if horizons:
        if "horizon" not in data.dims:
            raise ValueError("Data must have a 'horizon' dimension to be subsetted.")
        data = data.sel(horizon=horizons)
    if "criteria" not in data.dims:
        data = ensembles.make_criteria(data)

    selected = getattr(ensembles, f"{method}_reduce_ensemble")(data=data, **kwargs)

    clusters = {}
    fig_data = {}
    if method == "kmeans":
        fig_data = selected[2]
        clusters_tmp = selected[1]
        selected = selected[0]
        realization = np.arange(len(clusters_tmp))

        clusters = {
            g: data.realization.isel(realization=realization[clusters_tmp == g])
            for g in np.unique(clusters_tmp)
        }
    selected = data.realization.isel(realization=selected)

    return selected, clusters, fig_data



import datetime
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import xarray as xr
import xclim as xc
from xclim.testing.helpers import test_timeseries as timeseries

import xscen as xs
from xscen.testing import datablock_3d


@pytest.mark.parametrize("suffix", [".zarr", ".zarr.zip", "h5", "nc"])
def test_get_engine(tmpdir, suffix):
    if suffix in [".zarr", ".zarr.zip"]:
        path = "some/path" + suffix
        assert xs.io.get_engine(path) == "zarr"
    else:
        ds = timeseries(
            np.zeros(60),
            variable="tas",
            as_dataset=True,
        )
        ds.to_netcdf(
            Path(tmpdir) / f"test.nc",
            engine="netcdf4" if suffix == "nc" else "h5netcdf",
        )
        assert xs.io.get_engine(Path(tmpdir) / f"test.nc") in [
            "netcdf4",
            "h5netcdf",
        ]  # Hard to predict which one


class TestEstimateChunks:
    ds = datablock_3d(
        np.zeros((50, 100, 150)),
        variable="tas",
        x="lon",
        x_start=-70,
        x_step=0.1,
        y="lat",
        y_start=45,
        y_step=-0.1,
        as_dataset=True,
    )
    ds2 = ds.copy()
    ds2["tas"] = ds2["tas"].astype(np.float32)
    ds["just_a_variable"] = xr.DataArray(np.zeros(50), dims="new_dim")

    def test_normal(self):
        out1 = xs.io.estimate_chunks(self.ds, dims=["time", "lat", "lon"], target_mb=1)
        assert out1 == {"time": 30, "lat": 55, "lon": 85}
        out2 = xs.io.estimate_chunks(self.ds2, dims=["time", "lat", "lon"], target_mb=1)
        assert out2 == {"time": 35, "lat": 70, "lon": 105}
        out3 = xs.io.estimate_chunks(self.ds, dims=["lat", "lon"], target_mb=1)
        assert out3 == {"lon": 65, "lat": 40, "time": -1}
        out4 = xs.io.estimate_chunks(self.ds2, dims=["time"], target_mb=1)
        assert out4 == {"time": 15, "lat": -1, "lon": -1}

    @pytest.mark.parametrize("chunk_per_variable", [True, False])
    @pytest.mark.parametrize("as_file", [True, False])
    def test_multiple_vars(self, tmpdir, chunk_per_variable, as_file):
        ds = self.ds.copy()
        ds["pr"] = ds["tas"].isel(time=0)

        if as_file:
            ds.to_netcdf(Path(tmpdir) / "test.nc")
            ds = Path(tmpdir) / "test.nc"

        out = xs.io.estimate_chunks(
            ds, dims=["lat", "lon"], target_mb=1, chunk_per_variable=chunk_per_variable
        )
        if chunk_per_variable is False:
            assert out == {"lon": 65, "lat": 40, "time": -1}
        else:
            assert out == {
                "tas": {"lon": 65, "lat": 40, "time": -1},
                "pr": {"lon": 150, "lat": 100},
            }


class TestSubsetMaxsize:
    def test_normal(self):
        ds = datablock_3d(
            np.zeros((1500, 5, 5)),
            variable="tas",
            x="lon",
            x_start=-70,
            x_step=0.1,
            y="lat",
            y_start=45,
            y_step=-0.1,
            as_dataset=True,
        )
        ds["pr"] = ds["tas"]
        # First, test with a dataset that is already small enough
        out = xs.io.subset_maxsize(ds, maxsize_gb=1)
        assert len(out) == 1
        assert out[0].equals(ds)

        out = xs.io.subset_maxsize(ds, maxsize_gb=0.0005)
        assert len(out) == 2
        assert xr.concat(out, dim="time").equals(ds)

    def test_error(self):
        ds = datablock_3d(
            np.zeros((1, 50, 10)),
            variable="tas",
            x="lon",
            x_start=-70,
            x_step=0.1,
            y="lat",
            y_start=45,
            y_step=-0.1,
            as_dataset=True,
        )
        ds = ds.isel(time=0)

        with pytest.raises(NotImplementedError, match="does not contain a"):
            xs.io.subset_maxsize(ds, maxsize_gb=1e-15)


class TestCleanIncomplete:
    @pytest.mark.parametrize("which", ["complete", "incomplete"])
    def test_complete(self, tmpdir, which):
        ds = datablock_3d(
            np.ones((5, 5, 5)),
            variable="tas",
            x="lon",
            x_start=-70,
            x_step=0.1,
            y="lat",
            y_start=45,
            y_step=-0.1,
            as_dataset=True,
        )
        ds["pr"] = ds["tas"].copy()
        ds.to_zarr(Path(tmpdir) / "test.zarr")

        if which == "complete":
            xs.io.clean_incomplete(Path(tmpdir) / "test.zarr", complete=["tas"])
        else:
            xs.io.clean_incomplete(Path(tmpdir) / "test.zarr", incomplete=["pr"])
        assert (Path(tmpdir) / "test.zarr/tas").exists()
        assert not (Path(tmpdir) / "test.zarr/pr").exists()
        assert (Path(tmpdir) / "test.zarr/.zmetadata").exists()

        ds2 = xr.open_zarr(Path(tmpdir) / "test.zarr")
        assert "pr" not in ds2
        assert ds2.equals(ds[["tas"]])

    def test_error(self, tmpdir):
        ds = datablock_3d(
            np.ones((5, 5, 5)),
            variable="tas",
            x="lon",
            x_start=-70,
            x_step=0.1,
            y="lat",
            y_start=45,
            y_step=-0.1,
            as_dataset=True,
        )
        ds["pr"] = ds["tas"].copy()
        ds.to_zarr(Path(tmpdir) / "test.zarr")

        with pytest.raises(ValueError, match="Use either"):
            xs.io.clean_incomplete(
                Path(tmpdir) / "test.zarr", complete=["tas"], incomplete=["pr"]
            )


class TestRechunkForSaving:
    @pytest.mark.parametrize(
        "dims, xy",
        [
            (["lon", "lat"], True),
            (["lon", "lat"], False),
            (["rlon", "rlat"], True),
            (["rlon", "rlat"], False),
        ],
    )
    def test_options(self, dims, xy):
        ds = datablock_3d(
            np.random.random((30, 30, 50)),
            variable="tas",
            x=dims[0],
            x_start=-70 if dims[0] == "lon" else 0,
            y=dims[1],
            y_start=45 if dims[1] == "lat" else 0,
            as_dataset=True,
        )
        x = "X" if xy else dims[0]
        y = "Y" if xy else dims[1]
        new_chunks = {y: 10, x: 10, "time": 20}
        ds_ch = xs.io.rechunk_for_saving(ds, new_chunks)
        for dim, chunks in ds_ch.chunks.items():
            dim = (
                dim
                if (xy is False or dim == "time")
                else "X" if dim == dims[0] else "Y"
            )
            assert chunks[0] == new_chunks[dim]

    def test_variables(self):
        ds = datablock_3d(
            np.random.random((30, 30, 50)),
            variable="tas",
            x="lon",
            x_start=-70,
            y="lat",
            y_start=45,
            as_dataset=True,
        )
        ds["pr"] = datablock_3d(
            np.random.random((30, 30, 50)),
            variable="pr",
            x="lon",
            x_start=-70,
            y="lat",
            y_start=45,
            as_dataset=False,
        )

        new_chunks = {
            "tas": {"time": 20, "lon": 10, "lat": 7},
            "pr": {"time": 10, "lon": 5, "lat": 3},
        }
        ds_ch = xs.io.rechunk_for_saving(ds, new_chunks)
        for v in ds_ch.data_vars:
            for dim, chunks in zip(list(ds.dims), ds_ch[v].chunks):
                assert chunks[0] == new_chunks[v][dim]


class TestToTable:
    ds = xs.utils.unstack_dates(
        xr.merge(
            [
                xs.testing.datablock_3d(
                    np.ones((20, 3, 2)),
                    v,
                    "lon",
                    0,
                    "lat",
                    0,
                    1,
                    1,
                    "1993-01-01",
                    "QS-JAN",
                )
                for v in ["tas", "pr", "snw"]
            ]
        )
        .stack(site=["lat", "lon"])
        .reset_index("site")
        .assign_coords(site=list("abcdef"))
    ).transpose("season", "time", "site")
    ds.attrs = {"foo": "bar", "baz": 1, "qux": 2.0}

    @pytest.mark.parametrize(
        "multiple, as_dataset", [(True, True), (False, True), (False, False)]
    )
    def test_normal(self, tmpdir, multiple, as_dataset):
        if multiple is False:
            if as_dataset:
                ds = self.ds[["tas"]].copy()
            else:
                ds = self.ds["tas"].copy()
        else:
            ds = self.ds.copy()

        # Default
        xs.save_to_table(ds, Path(tmpdir) / "test.csv")
        saved = pd.read_csv(Path(tmpdir) / "test.csv")
        tab = xs.io.to_table(ds)

        assert tab.shape == (
            120,
            5 if multiple else 3,
        )  # 3 variables + 2 coords that are not dimensions
        assert saved.shape == (
            120,
            8 if multiple else 6,
        )  # everything gets mapped, so dimensions are included in the columns
        assert tab.columns.names == ["variable"] if multiple else [None]
        assert (
            set(saved.columns)
            == {"season", "time", "site", "lat", "lon", "pr", "snw", "tas"}
            if multiple
            else {"season", "time", "site", "tas"}
        )
        assert tab.index.names == ["season", "time", "site"]
        # Season order is chronological, rather than alphabetical
        np.testing.assert_array_equal(
            tab.xs("1993", level="time")
            .xs("a", level="site")
            .index.get_level_values("season"),
            ["JFM", "AMJ", "JAS", "OND"],
        )
        np.testing.assert_array_equal(saved.loc[0, "season"], "JFM")

        if multiple:
            # Variable in the index, thus no coords
            xs.save_to_table(
                ds,
                Path(tmpdir) / "test.xlsx",
                row=["time", "variable"],
                column=["season", "site"],
                coords=False,
            )
            tab = xs.io.to_table(
                ds, row=["time", "variable"], column=["season", "site"], coords=False
            )
            saved = pd.read_excel(Path(tmpdir) / "test.xlsx")

            assert tab.shape == (15, 24)
            assert saved.shape == (17, 26)  # Because of the headers
            assert tab.columns.names == ["season", "site"]
            np.testing.assert_array_equal(
                tab.loc[("1993", "pr"), ("JFM",)], ds.pr.sel(time="1993", season="JFM")
            )
            # Ensure that the coords are not present
            assert (
                len(
                    set(tab.index.get_level_values("variable").unique()).difference(
                        ["tas", "pr", "snw"]
                    )
                )
                == 0
            )
            # Excel is not the prettiest thing to test
            np.testing.assert_array_equal(saved.iloc[2, 2:], np.tile([1], 24))
            assert saved.iloc[0, 2] == "a"
            assert saved.iloc[2, 0] == datetime.datetime(1993, 1, 1, 0, 0)

    def test_sheet(self, tmpdir):
        xs.save_to_table(
            self.ds,
            Path(tmpdir) / "test.xlsx",
            row=["time", "variable"],
            column=["season"],
            sheet="site",
            coords=False,
        )
        saved = pd.read_excel(
            Path(tmpdir) / "test.xlsx", sheet_name=["a", "b", "c", "d", "e", "f"]
        )  # This is a test by itself
        tab = xs.io.to_table(
            self.ds,
            row=["time", "variable"],
            column=["season"],
            sheet="site",
            coords=False,
        )

        assert set(tab.keys()) == {("a",), ("b",), ("c",), ("d",), ("e",), ("f",)}
        assert tab[("a",)].shape == (15, 4)  # 5 time * 3 variable X 4 season
        assert saved["a"].shape == (15, 6)  # Because of the headers

    def test_kwargs(self, tmpdir):
        xs.save_to_table(
            self.ds,
            Path(tmpdir) / "test.xlsx",
            row=["time", "variable"],
            column=["season", "site"],
            coords=False,
            datetime_format="dd/mm/yyyy",
        )
        saved = pd.read_excel(Path(tmpdir) / "test.xlsx")
        assert saved.iloc[2, 0] == datetime.datetime(
            1993, 1, 1, 0, 0
        )  # No real way to test the format

    def test_multiindex(self, tmpdir):
        xs.save_to_table(
            self.ds,
            Path(tmpdir) / "test.csv",
            row=["time", "variable"],
            column=["season", "site"],
            coords=False,
            row_sep="|",
            col_sep=";",
        )
        out = pd.read_csv(Path(tmpdir) / "test.csv")
        assert out.shape == (15, 25)
        assert out.columns[0] == "time|variable"
        assert out.columns[1] == "JFM;a"

    def test_error(self, tmpdir):
        with pytest.raises(ValueError, match="Repeated dimension names."):
            xs.save_to_table(
                self.ds,
                Path(tmpdir) / "test.xlsx",
                row=["time", "variable"],
                column=["season", "site", "time"],
            )
        with pytest.raises(ValueError, match="Passed row, column and sheet"):
            xs.save_to_table(
                self.ds,
                Path(tmpdir) / "test.xlsx",
                row=["time", "variable"],
                column=["season", "site", "foo"],
            )
        with pytest.raises(
            NotImplementedError,
            match="Keeping auxiliary coords is not implemented when",
        ):
            xs.save_to_table(
                self.ds,
                Path(tmpdir) / "test.xlsx",
                row=["time", "variable"],
                column=["season", "site"],
                coords=True,
            )
        with pytest.raises(ValueError, match="Output format could not be inferred"):
            xs.save_to_table(self.ds, Path(tmpdir) / "test")
        with pytest.raises(
            ValueError, match="is only valid with excel as the output format"
        ):
            xs.save_to_table(self.ds, Path(tmpdir) / "test.csv", sheet="site")
        with pytest.raises(ValueError, match="but the output format is not Excel."):
            xs.save_to_table(self.ds, Path(tmpdir) / "test.csv", add_toc=True)

    @pytest.mark.parametrize("as_dataset", [True, False])
    def test_make_toc(self, tmpdir, as_dataset):
        ds = self.ds.copy()
        for v in ds.data_vars:
            ds[v].attrs["long_name"] = f"Long name for {v}"
            ds[v].attrs["long_name_fr"] = f"Nom long pour {v}"

        if as_dataset is False:
            ds = ds["tas"]

        with xc.set_options(metadata_locales="fr"):
            xs.save_to_table(ds, Path(tmpdir) / "test.xlsx", add_toc=True)

        toc = pd.read_excel(Path(tmpdir) / "test.xlsx", sheet_name="Contenu")
        toc = toc.set_index("Unnamed: 0" if as_dataset else "Variable")

        if as_dataset:
            assert toc.shape == (8, 2)
            assert toc.columns.tolist() == ["Description", "Unités"]
            assert toc.index.tolist() == [
                "tas",
                "pr",
                "snw",
                np.nan,
                "Attributs globaux",
                "foo",
                "baz",
                "qux",
            ]
            assert toc.loc["tas", "Description"] == "Nom long pour tas"
            assert toc.loc["tas", "Unités"] == "K"
        else:
            assert toc.shape == (1, 2)
            assert toc.columns.tolist() == ["Description", "Unités"]
            assert toc.index.tolist() == ["tas"]
            assert toc.loc["tas", "Description"] == "Nom long pour tas"
            assert toc.loc["tas", "Unités"] == "K"


def test_round_bits():
    da = datablock_3d(
        np.random.random((30, 30, 50)),
        variable="tas",
        x="lon",
        x_start=-70,
        y="lat",
        y_start=45,
    )
    dar = xs.io.round_bits(da, 12)
    # Close but NOT equal, meaning something happened
    np.testing.assert_allclose(da, dar, rtol=0.013)
    # There's always a chance of having a randomly chosen number with only zeros in the bit rounded part of the mantissa
    # Assuming a uniform distribution of binary numbers (which it is not), the chance of this happening should be:
    # 2^(23 - 12 + 1) / 2^24 = 2^(-12) ~ 0.02 % (but we'll allow 1% of values to be safe)
    assert (da != dar).sum() > (0.99 * da.size)


class TestSaveToZarr:
    @pytest.mark.parametrize(
        "vname,vtype,bitr,exp",
        [
            ("tas", np.float32, 12, 12),
            ("tas", np.float32, False, None),
            ("tas", np.int32, 12, None),
            ("tas", np.int32, {"tas": 2}, "error"),
            ("tas", object, {"pr": 2}, None),
            ("tas", np.float64, True, 12),
        ],
    )
    def test_guess_bitround(self, vname, vtype, bitr, exp):
        if exp == "error":
            with pytest.raises(ValueError):
                xs.io._get_keepbits(bitr, vname, vtype)
        else:
            assert xs.io._get_keepbits(bitr, vname, vtype) == exp

    @pytest.mark.parametrize("mode", ["f", "o", "a"])
    @pytest.mark.parametrize("itervar", [True, False])
    def test_mode(self, tmpdir, mode, itervar):
        ds1 = timeseries(
            np.arange(1, 5),
            variable="tas",
            as_dataset=True,
        )
        xs.save_to_zarr(ds1, Path(tmpdir) / "test.zarr")

        ds2 = timeseries(
            np.arange(10, 14),
            variable="tas",
            as_dataset=True,
        )
        ds2["pr"] = ds2["tas"].copy()
        ds2 = ds2[["pr", "tas"]]

        if mode == "f":
            with pytest.raises(ValueError, match="exists in dataset"):
                xs.save_to_zarr(
                    ds2, Path(tmpdir) / "test.zarr", mode=mode, itervar=itervar
                )
            assert not (Path(tmpdir) / "test.zarr/pr").exists()
            if itervar:
                # Essentially just to reach 100% coverage and make sure the function doesn't crash with mode="f" and itervar=True
                xs.save_to_zarr(
                    ds2, Path(tmpdir) / "test2.zarr", mode=mode, itervar=itervar
                )
                ds3 = xr.open_zarr(Path(tmpdir) / "test2.zarr")
                np.testing.assert_array_almost_equal(ds3.tas.isel(time=0), [10])
                np.testing.assert_array_almost_equal(ds3.pr.isel(time=0), [10])

        elif mode == "o":
            xs.save_to_zarr(ds2, Path(tmpdir) / "test.zarr", mode=mode, itervar=itervar)
            ds3 = xr.open_zarr(Path(tmpdir) / "test.zarr")
            np.testing.assert_array_almost_equal(ds3.tas.isel(time=0), [10])
            np.testing.assert_array_almost_equal(ds3.pr.isel(time=0), [10])

        elif mode == "a":
            # First, try only with variables that are already in the dataset
            xs.save_to_zarr(
                ds2[["tas"]], Path(tmpdir) / "test.zarr", mode=mode, itervar=itervar
            )
            ds3 = xr.open_zarr(Path(tmpdir) / "test.zarr")
            np.testing.assert_array_almost_equal(ds3.tas.isel(time=0), [1])

            # Now, try with a new variable
            xs.save_to_zarr(ds2, Path(tmpdir) / "test.zarr", mode=mode, itervar=itervar)
            ds3 = xr.open_zarr(Path(tmpdir) / "test.zarr")
            np.testing.assert_array_almost_equal(ds3.tas.isel(time=0), [1])
            np.testing.assert_array_almost_equal(ds3.pr.isel(time=0), [10])

    @pytest.mark.parametrize("append", [True, False])
    def test_append(self, tmpdir, append):
        ds1 = datablock_3d(
            np.array([[[1, 2], [3, 4]]]),
            variable="tas",
            x="lon",
            x_start=-70,
            y="lat",
            y_start=45,
            as_dataset=True,
        )
        ds2 = datablock_3d(
            np.array([[[11, 12], [13, 14]]]),
            variable="tas",
            x="lon",
            x_start=-70,
            y="lat",
            y_start=45,
            start="2005-01-01",
            as_dataset=True,
        )
        ds2["pr"] = ds2["tas"].copy()
        xs.save_to_zarr(
            ds1, Path(tmpdir) / "test.zarr", encoding={"tas": {"dtype": "float32"}}
        )

        encoding = {
            "tas": {"dtype": "int32"}
        }  # This should be ignored, as the variable is already in the dataset
        if append:
            with pytest.raises(
                ValueError,
                match="is set in zarr_kwargs, all variables must already exist in the dataset.",
            ):
                xs.save_to_zarr(
                    ds2,
                    Path(tmpdir) / "test.zarr",
                    mode="a",
                    zarr_kwargs={"append_dim": "time"},
                    encoding=encoding,
                )
            xs.save_to_zarr(
                ds2[["tas"]],
                Path(tmpdir) / "test.zarr",
                mode="a",
                zarr_kwargs={"append_dim": "time"},
                encoding=encoding,
            )
            out = xr.open_zarr(Path(tmpdir) / "test.zarr")
            np.testing.assert_array_equal(
                out.tas, np.array([[[1, 2], [3, 4]], [[11, 12], [13, 14]]])
            )
        else:
            xs.save_to_zarr(
                ds2, Path(tmpdir) / "test.zarr", mode="a", encoding=encoding
            )
            out = xr.open_zarr(Path(tmpdir) / "test.zarr")
            np.testing.assert_array_equal(out.tas, np.array([[[1, 2], [3, 4]]]))
            np.testing.assert_array_equal(out.pr, np.array([[[11, 12], [13, 14]]]))
        assert out.tas.dtype == np.float32

    def test_skip(self, tmpdir):
        ds1 = timeseries(
            np.arange(1, 5),
            variable="tas",
            as_dataset=True,
        )
        ds2 = timeseries(
            np.arange(10, 14),
            variable="tas",
            as_dataset=True,
        )
        xs.save_to_zarr(ds1, Path(tmpdir) / "test.zarr")
        xs.save_to_zarr(ds2, Path(tmpdir) / "test.zarr", mode="a")
        ds3 = xr.open_zarr(Path(tmpdir) / "test.zarr")
        np.testing.assert_array_almost_equal(ds3.tas.isel(time=0), [1])


@pytest.mark.parametrize("engine", ["netcdf", "zarr"])
def test_savefuncs_normal(tmpdir, engine):
    ds = datablock_3d(
        np.tile(np.arange(1111, 1121), 15).reshape(15, 5, 2) * 1e-7,
        variable="tas",
        x="lon",
        x_start=-70,
        y="lat",
        y_start=45,
        as_dataset=True,
    )
    ds["pr"] = ds["tas"].copy()
    ds["other"] = ds["tas"].copy()
    ds["other"].encoding = {"dtype": "float32"}
    ds.attrs["foo"] = {"bar": 1}
    ds["pr"].attrs["foo"] = {"bar": 2}

    ds = ds.assign_coords(
        some_coord=("lat", np.array(["hi", "how", "are", "you", "doing"]))
    )
    ds["some_coord"] = ds["some_coord"].astype(object)
    ds["some_coord"].encoding = {"source": "this is a source"}

    rechunk = {"time": 5, "lon": 2, "lat": 2}
    bitround = {"tas": 2, "pr": 3}
    if engine == "netcdf":
        xs.save_to_netcdf(
            ds,
            Path(tmpdir) / "test.nc",
            rechunk=rechunk,
            bitround=bitround,
        )
        ds2 = xr.open_dataset(Path(tmpdir) / "test.nc", chunks={})
    else:
        xs.save_to_zarr(
            ds,
            Path(tmpdir) / "test.zarr",
            rechunk=rechunk,
            bitround=bitround,
        )
        ds2 = xr.open_zarr(Path(tmpdir) / "test.zarr")

    # Chunks
    assert ds2.tas.chunks == ((5, 5, 5), (2, 2, 1), (2,))

    # Dtype
    assert ds2.tas.dtype == np.float64
    assert ds2.other.dtype == np.float32

    # Bitround
    np.testing.assert_array_almost_equal(
        ds2.tas.isel(time=0, lat=0, lon=0), [0.00010681], decimal=8
    )
    assert ds2.tas.attrs["_QuantizeBitRoundNumberOfSignificantDigits"] == 2
    np.testing.assert_array_almost_equal(
        ds2.pr.isel(time=0, lat=0, lon=0), [0.00011444], decimal=8
    )
    assert ds2.pr.attrs["_QuantizeBitRoundNumberOfSignificantDigits"] == 3
    np.testing.assert_array_almost_equal(
        ds2.other.isel(time=0, lat=0, lon=0), [0.0001111], decimal=8
    )
    assert ds2.other.attrs["_QuantizeBitRoundNumberOfSignificantDigits"] == 12

    # Attributes
    assert ds2.attrs["foo"] == "{'bar': 1}"
    assert ds2.pr.attrs["foo"] == "{'bar': 2}"

    if engine == "netcdf":
        assert ds.some_coord.encoding == {"source": "this is a source"}
    else:
        assert ds.some_coord.encoding == {}


class TestRechunk:
    @pytest.mark.parametrize("engine", ["nc", "zarr"])
    def test_rechunk(self, tmpdir, engine):
        ds = datablock_3d(
            np.tile(np.arange(1111, 1121), 15).reshape(15, 5, 2) * 1e-7,
            variable="tas",
            x="lon",
            x_start=-70,
            y="lat",
            y_start=45,
            as_dataset=True,
        )
        ds["pr"] = ds["tas"].copy()

        if engine == "nc":
            xs.save_to_netcdf(
                ds,
                Path(tmpdir) / "test.nc",
            )
        else:
            xs.save_to_zarr(
                ds,
                Path(tmpdir) / "test.zarr",
            )

        (Path(tmpdir) / f"test2.zarr").mkdir()

        xs.io.rechunk(
            Path(tmpdir) / f"test.{engine}",
            Path(tmpdir) / "test2.zarr",
            chunks_over_dim={"time": 5, "lon": 2, "lat": 2},
            overwrite=True,
            worker_mem="1GB",
            temp_store=Path(tmpdir) / "temp",
        )
        xs.io.rechunk(
            Path(tmpdir) / f"test.{engine}",
            Path(tmpdir) / "test3.zarr",
            chunks_over_var={"tas": {"time": 5, "lon": 2, "lat": 2}},
            overwrite=True,
            worker_mem="1GB",
            temp_store=Path(tmpdir) / "temp",
        )

        ds2 = xr.open_zarr(Path(tmpdir) / "test2.zarr")
        ds3 = xr.open_zarr(Path(tmpdir) / "test3.zarr")
        assert ds2.tas.chunks == ((5, 5, 5), (2, 2, 1), (2,))
        assert ds2.pr.chunks == ((5, 5, 5), (2, 2, 1), (2,))
        assert ds3.tas.chunks == ((5, 5, 5), (2, 2, 1), (2,))
        assert ds3.pr.chunks == ((15,), (5,), (2,))

    def test_error(self, tmpdir):
        ds = datablock_3d(
            np.tile(np.arange(1111, 1121), 15).reshape(15, 5, 2) * 1e-7,
            variable="tas",
            x="lon",
            x_start=-70,
            y="lat",
            y_start=45,
            as_dataset=True,
        )
        with pytest.raises(ValueError, match="No chunks given. "):
            xs.io.rechunk(ds, Path(tmpdir) / "test.nc", worker_mem="1GB")


def test_zip_zip(tmpdir):
    ds = datablock_3d(
        np.tile(np.arange(1111, 1121), 15).reshape(15, 5, 2) * 1e-7,
        variable="tas",
        x="lon",
        x_start=-70,
        y="lat",
        y_start=45,
        as_dataset=True,
    )
    xs.save_to_zarr(ds, Path(tmpdir) / "test.zarr")
    xs.io.zip_directory(
        Path(tmpdir) / "test.zarr", Path(tmpdir) / "test.zarr.zip", delete=True
    )
    assert not (Path(tmpdir) / "test.zarr").exists()

    with xr.open_zarr(Path(tmpdir) / "test.zarr.zip") as ds2:
        assert ds2.equals(ds)

    xs.io.unzip_directory(Path(tmpdir) / "test.zarr.zip", Path(tmpdir) / "test2.zarr")
    with xr.open_zarr(Path(tmpdir) / "test2.zarr") as ds3:
        assert ds3.equals(ds)



"""Functions to compute xclim indicators."""

import logging
import os
from collections.abc import Sequence
from functools import partial
from pathlib import Path
from types import ModuleType

import pandas as pd
import xarray as xr
import xclim as xc
from intake_esm import DerivedVariableRegistry
from xclim.core.calendar import construct_offset, parse_offset
from xclim.core.indicator import Indicator
from yaml import safe_load

from xscen.config import parse_config

from .catutils import parse_from_ds
from .utils import CV, rechunk_for_resample, standardize_periods

logger = logging.getLogger(__name__)

__all__ = ["compute_indicators", "load_xclim_module", "registry_from_module"]


def load_xclim_module(filename: str | os.PathLike, reload: bool = False) -> ModuleType:
    """Return the xclim module described by the yaml file (or group of yaml, jsons and py).

    Parameters
    ----------
    filename : str or os.PathLike
        The filepath to the yaml file of the module or to the stem of yaml, jsons and py files.
    reload : bool
        If False (default) and the module already exists in `xclim.indicators`, it is not re-build.

    Returns
    -------
    ModuleType
        The xclim module.
    """
    if not reload:
        # Same code as in xclim to get the module name.
        filepath = Path(filename)

        if not filepath.suffix:
            # A stem was passed, try to load files
            ymlpath = filepath.with_suffix(".yml")
        else:
            ymlpath = filepath

        # Read YAML file
        with ymlpath.open() as f:
            yml = safe_load(f)

        name = yml.get("module", filepath.stem)
        if hasattr(xc.indicators, name):
            return getattr(xc.indicators, name)

    return xc.build_indicator_module_from_yaml(filename)


def get_indicator_outputs(ind: xc.core.indicator.Indicator, in_freq: str):
    """Returns the variables names and resampling frequency of a given indicator.

    CAUTION : Some indicators will build the variable name on-the-fly according to the arguments.
    This function will return the template string (with "{}").

    Parameters
    ----------
    ind : Indicator
        An xclim indicator
    in_freq : str
        The data's sampling frequency.

    Returns
    -------
    var_names : list
        List of variable names
    freq : str
        Indicator resampling frequency. "fx" for time-reducing indicator.
    """
    if isinstance(ind, xc.core.indicator.ReducingIndicator):
        frq = "fx"
    elif not isinstance(ind, xc.core.indicator.ResamplingIndicator):
        frq = in_freq
    else:
        frq = (
            ind.injected_parameters["freq"]
            if "freq" in ind.injected_parameters
            else ind.parameters["freq"].default
        )
    if frq == "YS":
        frq = "YS-JAN"
    var_names = [cfa["var_name"] for cfa in ind.cf_attrs]
    return var_names, frq


@parse_config
def compute_indicators(  # noqa: C901
    ds: xr.Dataset,
    indicators: (
        str
        | os.PathLike
        | Sequence[Indicator]
        | Sequence[tuple[str, Indicator]]
        | ModuleType
    ),
    *,
    periods: list[str] | list[list[str]] | None = None,
    restrict_years: bool = True,
    to_level: str | None = "indicators",
    rechunk_input: bool = False,
) -> dict:
    """Calculate variables and indicators based on a YAML call to xclim.

    The function cuts the output to be the same years as the inputs.
    Hence, if an indicator creates a timestep outside the original year range (e.g. the first DJF for QS-DEC),
    it will not appear in the output.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to use for the indicators.
    indicators : str | os.PathLike | Sequence[Indicator] | Sequence[tuple[str, Indicator]] | ModuleType
        Path to a YAML file that instructs on how to calculate missing variables.
        Can also be only the "stem", if translations and custom indices are implemented.
        Can be the indicator module directly, or a sequence of indicators or a sequence of
        tuples (indicator name, indicator) as returned by `iter_indicators()`.
    periods : list of str or list of lists of str, optional
        Either [start, end] or list of [start, end] of continuous periods over which to compute the indicators.
        This is needed when the time axis of ds contains some jumps in time.
        If None, the dataset will be considered continuous.
    restrict_years : bool
        If True, cut the time axis to be within the same years as the input.
        This is mostly useful for frequencies that do not start in January, such as QS-DEC.
        In that instance, `xclim` would start on previous_year-12-01 (DJF), with a NaN.
        `restrict_years` will cut that first timestep. This should have no effect on YS and MS indicators.
    to_level : str, optional
        The processing level to assign to the output.
        If None, the processing level of the inputs is preserved.
    rechunk_input : bool
        If True, the dataset is rechunked with :py:func:`flox.xarray.rechunk_for_blockwise`
        according to the resampling frequency of the indicator. Each rechunking is done
        once per frequency with :py:func:`xscen.utils.rechunk_for_resample`.

    Returns
    -------
    dict
        Dictionary (keys = timedeltas) with indicators separated by temporal resolution.

    See Also
    --------
    xclim.indicators, xclim.core.indicator.build_indicator_module_from_yaml
    """
    if isinstance(indicators, str | os.PathLike):
        logger.debug("Loading indicator module.")
        module = load_xclim_module(indicators)
        indicators = module.iter_indicators()
    elif hasattr(indicators, "iter_indicators"):
        indicators = indicators.iter_indicators()

    try:
        N = len(indicators)
    except TypeError:
        N = None
    else:
        msg = f"Computing {N} indicators."
        logger.info(msg)

    periods = standardize_periods(periods)
    in_freq = xr.infer_freq(ds.time) if "time" in ds.dims else "fx"
    dss_rechunked = {}

    out_dict = dict()
    for i, ind in enumerate(indicators, 1):
        if isinstance(ind, tuple):
            iden, ind = ind
        else:
            iden = ind.identifier
        msg = f"{i} - Computing {iden}."
        logger.info(msg)

        _, freq = get_indicator_outputs(ind, in_freq)

        if rechunk_input and freq not in ["fx", in_freq]:
            if freq not in dss_rechunked:
                msg = f"Rechunking with flox for freq {freq}."
                logger.debug(msg)
                dss_rechunked[freq] = rechunk_for_resample(ds, time=freq)
            else:
                msg = f"Using rechunked for freq {freq}"
                logger.debug(msg)
            ds_in = dss_rechunked[freq]
        else:
            ds_in = ds

        if periods is None:
            # Pandas as no semiannual frequency and 2Q is capricious
            if freq.startswith("2Q"):
                logger.debug(
                    "Dropping start of timeseries to ensure semiannual frequency works."
                )
                ds_in = fix_semiannual(ds_in, freq)
            # Make the call to xclim
            out = ind(ds=ds_in)

            # In the case of multiple outputs, merge them into a single dataset
            if isinstance(out, tuple):
                out = xr.merge(out)
                out.attrs = {}
            else:
                out = out.to_dataset()

        else:
            # Multiple time periods to concatenate
            concats = []
            for period in periods:
                # Make the call to xclim
                ds_subset = ds_in.sel(time=slice(period[0], period[1]))
                # Pandas as no semiannual frequency and 2Q is capricious
                if freq.startswith("2Q"):
                    logger.debug(
                        "Dropping start of timeseries to ensure semiannual frequency works."
                    )
                    ds_subset = fix_semiannual(ds_subset, freq)
                tmp = ind(ds=ds_subset)

                # In the case of multiple outputs, merge them into a single dataset
                if isinstance(tmp, tuple):
                    tmp = xr.merge(tmp)
                    tmp.attrs = {}
                else:
                    tmp = tmp.to_dataset()

                # In order to concatenate time periods, the indicator still needs a time dimension
                if freq == "fx":
                    tmp = tmp.assign_coords({"time": ds_subset.time[0]})
                concats.append(tmp)
            out = xr.concat(concats, dim="time")

        # Make sure that attributes have been kept for the dimensions and coordinates. Fixes a bug in xarray.
        for c in set(list(out.coords) + list(out.dims)).intersection(
            set(list(ds.coords) + list(ds.dims))
        ):
            if (out[c].attrs != ds[c].attrs) and (out[c].sizes == ds[c].sizes):
                out[c].attrs = ds[c].attrs

        if restrict_years and "time" in out.dims:
            # cut the time axis to be within the same years as the input
            # for QS-DEC, xclim starts on DJF with time previous_year-12-01 with a nan as values. We want to cut this.
            # this should have no effect on YS and MS indicators
            out = out.sel(
                time=slice(
                    str(ds.time[0].dt.year.values), str(ds.time[-1].dt.year.values)
                )
            )

        # Create the dictionary key
        key = freq
        if key not in out_dict:
            out_dict[key] = out
            out_dict[key].attrs = ds.attrs
            out_dict[key].attrs["cat:xrfreq"] = freq
            out_dict[key].attrs["cat:frequency"] = CV.xrfreq_to_frequency(freq, None)
            if to_level is not None:
                out_dict[key].attrs["cat:processing_level"] = to_level

        else:
            for v in out.data_vars:
                out_dict[key][v] = out[v]
    for key in out_dict:
        out_dict[key].attrs["cat:variable"] = parse_from_ds(
            out_dict[key], ["variable"]
        )["variable"]

    return out_dict


def registry_from_module(
    module: ModuleType,
    registry: DerivedVariableRegistry | None = None,
    variable_column: str = "variable",
) -> DerivedVariableRegistry:
    """Convert a xclim virtual indicators module to an intake_esm Derived Variable Registry.

    Parameters
    ----------
    module : ModuleType
        A module of xclim.
    registry : DerivedVariableRegistry, optional
        If given, this registry is extended, instead of creating a new one.
    variable_column : str
        The name of the variable column (the name used in the query).

    Returns
    -------
    DerivedVariableRegistry
        A variable registry where each indicator and each of its output has been registered.
        If an indicator returns multiple values, each of them is mapped individually, as
        the DerivedVariableRegistry only supports single output function.
        Each indicator was wrapped into a new function that only accepts a dataset and
        returns it with the extra variable appended. This means all other parameters are
        given their defaults.
    """
    dvr = registry or DerivedVariableRegistry()
    for name, ind in module.iter_indicators():
        query = {
            variable_column: [p.default for p in ind.parameters.values() if p.kind == 0]
        }
        for i, attrs in enumerate(ind.cf_attrs):
            dvr.register(variable=attrs["var_name"], query=query)(_derived_func(ind, i))
    return dvr


def _ensure_list(x):
    if not isinstance(x, list | tuple):
        return [x]
    return x


def _derived_func(ind: xc.core.indicator.Indicator, nout: int) -> partial:
    def func(ds, *, ind, nout):
        out = ind(ds=ds)
        if isinstance(out, tuple):
            out = out[nout]
        ds[out.name] = out
        return ds

    func.__name__ = ind.identifier
    return partial(func, ind=ind, nout=nout)


def select_inds_for_avail_vars(
    ds: xr.Dataset,
    indicators: (
        str
        | os.PathLike
        | Sequence[Indicator]
        | Sequence[tuple[str, Indicator]]
        | ModuleType
    ),
) -> ModuleType:
    """Filter the indicators for which the necessary variables are available.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to use for the indicators.
    indicators : str | os.PathLike | Sequence[Indicator] | Sequence[Tuple[str, Indicator]]
        Path to a YAML file that instructs on how to calculate indicators.
        Can also be only the "stem", if translations and custom indices are implemented.
        Can be the indicator module directly, or a sequence of indicators or a sequence of
        tuples (indicator name, indicator) as returned by `iter_indicators()`.

    Returns
    -------
    ModuleType – An indicator module of 'length' ∈ [0, n].

    See Also
    --------
    xclim.indicators, xclim.core.indicator.build_indicator_module_from_yaml
    """
    # Transform the 'indicators' input into a list of tuples (name, indicator)
    is_list_of_tuples = isinstance(indicators, list) and all(
        isinstance(i, tuple) for i in indicators
    )
    if isinstance(indicators, str | os.PathLike):
        logger.debug("Loading indicator module.")
        indicators = load_xclim_module(indicators, reload=True)
    if hasattr(indicators, "iter_indicators"):
        indicators = [(name, ind) for name, ind in indicators.iter_indicators()]
    elif isinstance(indicators, list | tuple) and not is_list_of_tuples:
        indicators = [(ind.base, ind) for ind in indicators]

    # FIXME: Remove if-else when updating minimum xclim version to 0.53
    XCVARS = (
        xc.core.VARIABLES if hasattr(xc.core, "VARIABLES") else xc.core.utils.VARIABLES
    )
    available_vars = {var for var in ds.data_vars if var in XCVARS.keys()}
    available_inds = [
        (name, ind)
        for var in available_vars
        for name, ind in indicators
        if var in ind.parameters.keys()
    ]
    return xc.core.indicator.build_indicator_module(
        "inds_for_avail_vars", available_inds, reload=True
    )


def _wrap_month(m):
    # Ensure the month number is between 1 and 12
    # Modulo returns 0 if m is a multiple of 12, 0 is false and we want 12.
    return (m % 12) or 12


def fix_semiannual(ds, freq):
    """
    Avoid wrong start dates for semiannual frequency.

    Resampling with offsets that are multiples of a base frequency (ex: 2QS-OCT) is broken in pandas (https://github.com/pandas-dev/pandas/issues/51563).
    This will cut the beginning of the dataset so it starts exactly at the beginning of the resampling period.
    """
    # I hate that we have to do that
    mul, b, s, anc = parse_offset(freq)
    if mul != 2 or b != "Q":
        raise NotImplementedError("This only fixes 2Q frequencies.")
    # Get MONTH: N mapping (invert xarray's)
    months_inv = xr.coding.cftime_offsets._MONTH_ABBREVIATIONS
    months = dict(zip(months_inv.values(), months_inv.keys()))

    if s:
        m1 = months[anc]
    else:
        m1 = _wrap_month(months[anc] + 1)
        freq = construct_offset(mul, b, True, months_inv[m1])
    m2 = _wrap_month(m1 + 6)

    time = ds.indexes["time"]
    if isinstance(time, xr.CFTimeIndex):
        offset = xr.coding.cftime_offsets.to_offset(freq)
        is_on_offset = offset.onOffset
    else:
        offset = pd.tseries.frequencies.to_offset(freq)
        is_on_offset = offset.is_on_offset

    if is_on_offset(time[0]) and time[0].month in (m1, m2):
        # wow, already correct!
        return ds

    for t in time:
        if is_on_offset(t) and t.month in (m1, m2):
            return ds.sel(time=(time >= t))
    raise ValueError(f"Can't find a start date that fits with frequency {freq}.")



import numpy as np
import pytest
import xarray as xr
from conftest import notebooks
from xclim.testing.helpers import test_timeseries as timeseries

import xscen as xs
from xscen.testing import datablock_3d


class TestHealthChecks:
    def test_structure(self, datablock_3d):
        ds = datablock_3d(
            np.random.rand(5, 5, 5), "tas", "rlon", 0, "rlat", 0, as_dataset=True
        )

        # Right structure
        xs.diagnostics.health_checks(
            ds,
            structure={
                "dims": ["time", "rlat", "rlon"],
                "coords": ["time", "rlat", "rlon", "lon", "lat", "rotated_pole"],
            },
            raise_on=["all"],
        )

        # Wrong structures
        with pytest.raises(ValueError, match="The dimension 'lat' is missing."):
            xs.diagnostics.health_checks(
                ds, structure={"dims": ["time", "lat", "rlon"]}, raise_on=["all"]
            )
        with pytest.warns(UserWarning, match="The dimension 'lat' is missing."):
            xs.diagnostics.health_checks(
                ds, structure={"dims": ["time", "lat", "rlon"]}
            )
        with pytest.raises(ValueError, match="The coordinate 'another' is missing."):
            xs.diagnostics.health_checks(
                ds, structure={"coords": ["another"]}, raise_on=["structure"]
            )
        with pytest.warns(UserWarning, match="The coordinate 'another' is missing."):
            xs.diagnostics.health_checks(ds, structure={"coords": ["another"]})
        with pytest.raises(
            ValueError, match="'tas' is detected as a data variable, not a coordinate."
        ):
            xs.diagnostics.health_checks(
                ds, structure={"coords": ["tas"]}, raise_on=["structure"]
            )
        with pytest.warns(
            UserWarning, match="'tas' is detected as a data variable, not a coordinate."
        ):
            xs.diagnostics.health_checks(ds, structure={"coords": ["tas"]})

    @pytest.mark.parametrize("cal", ["default", "standard", "365_day"])
    def test_calendar(self, cal):
        ds = timeseries(np.arange(0, 365), "tas", "1/1/2000", freq="D", as_dataset=True)
        if cal != "default":
            ds = ds.convert_calendar(cal)

        calendars = ["standard", "proleptic_gregorian", "noleap", "365_day", "360_day"]
        fail_on = [0, 1, 4] if cal == "365_day" else [2, 3, 4]
        for i, cals in enumerate(calendars):
            if i in fail_on:
                with pytest.raises(ValueError, match=f"The calendar is not '{cals}'."):
                    xs.diagnostics.health_checks(ds, calendar=cals, raise_on=["all"])
                with pytest.warns(UserWarning, match=f"The calendar is not '{cals}'."):
                    xs.diagnostics.health_checks(ds, calendar=cals)
            else:
                xs.diagnostics.health_checks(ds, calendar=cals, raise_on=["all"])

    @pytest.mark.parametrize("cal", ["default", "360_day", "365_day"])
    def test_dates(self, cal):
        ds = timeseries(np.arange(0, 365), "tas", "1/1/2000", freq="D", as_dataset=True)
        if cal != "default":
            ds = ds.convert_calendar(cal, align_on="date")

        # Right dates
        xs.diagnostics.health_checks(
            ds, start_date="2000-01-01", end_date="2000-12-30", raise_on=["all"]
        )
        xs.diagnostics.health_checks(ds, start_date="2000-01-02", raise_on=["all"])
        xs.diagnostics.health_checks(ds, end_date="2000-01-02", raise_on=["all"])

        # Wrong dates
        with pytest.raises(
            ValueError, match="The start date is not at least 1999-01-02."
        ):
            xs.diagnostics.health_checks(ds, start_date="1999-01-02", raise_on=["all"])
        with pytest.warns(
            UserWarning, match="The start date is not at least 1999-01-02."
        ):
            xs.diagnostics.health_checks(ds, start_date="1999-01-02")
        with pytest.raises(
            ValueError, match="The end date is not at least 2001-01-01."
        ):
            xs.diagnostics.health_checks(
                ds, end_date="2001-01-01", raise_on=["end_date"]
            )
        with pytest.warns(
            UserWarning, match="The end date is not at least 2001-01-01."
        ):
            xs.diagnostics.health_checks(ds, end_date="2001-01-01")

    def test_variables(self):
        ds = timeseries(np.arange(0, 365), "tas", "1/1/2000", freq="D")

        # Right units
        xs.diagnostics.health_checks(
            ds, variables_and_units={"tas": "K"}, raise_on=["all"]
        )

        # Wrong variables or units
        with pytest.raises(ValueError, match="The variable 'tas2' is missing."):
            xs.diagnostics.health_checks(
                ds, variables_and_units={"tas2": "K"}, raise_on=["all"]
            )
        with pytest.raises(
            ValueError,
            match="The variable 'tas' does not have the expected units 'degC'. Received 'K'.",
        ):
            xs.diagnostics.health_checks(
                ds, variables_and_units={"tas": "degC"}, raise_on=["all"]
            )
        with pytest.warns(
            UserWarning,
            match="The variable 'tas' does not have the expected units 'degC'. Received 'K'.",
        ):
            xs.diagnostics.health_checks(ds, variables_and_units={"tas": "degC"})
        with pytest.warns(
            UserWarning,
            match="are not compatible with requested mm.",
        ):
            xs.diagnostics.health_checks(ds, variables_and_units={"tas": "mm"})

    def test_cfchecks(self):
        ds = timeseries(np.arange(0, 365), "tas", "1/1/2000", freq="D", as_dataset=True)

        # Check that the cfchecks are valid
        with pytest.raises(ValueError, match="Check 'fake' is not in xclim."):
            xs.diagnostics.health_checks(ds, cfchecks={"tas": {"fake": {}}})

        # Good checks
        cfcheck = {
            "tas": {
                "check_valid": {"key": "standard_name", "expected": "air_temperature"},
                "cfcheck_from_name": {},
            }
        }
        xs.diagnostics.health_checks(ds, cfchecks=cfcheck, raise_on=["all"])

        # Bad checks
        bad_cfcheck = {
            "tas": {
                "check_valid": {"key": "standard_name", "expected": "something_else"},
                "cfcheck_from_name": {"varname": "pr"},
            }
        }
        with pytest.raises(
            ValueError, match="['something_else']"
        ):  # Will raise on the first check
            xs.diagnostics.health_checks(ds, cfchecks=bad_cfcheck, raise_on=["all"])
        with pytest.warns(
            UserWarning, match="['precipitation_flux']"
        ):  # Make sure the second check is still run
            xs.diagnostics.health_checks(ds, cfchecks=bad_cfcheck)

    @pytest.mark.parametrize(
        "freq, gap", [("D", False), ("MS", False), ("3h", False), ("D", True)]
    )
    def test_freq(self, freq, gap):
        ds = timeseries(
            np.arange(0, 365), "tas", "1/1/2000", freq=freq, as_dataset=True
        )
        if gap is False:
            # Right frequency
            xs.diagnostics.health_checks(ds, freq=freq, raise_on=["all"])

            # Wrong frequency
            with pytest.raises(ValueError, match="The frequency is not 'M'."):
                xs.diagnostics.health_checks(ds, freq="M", raise_on=["all"])
            with pytest.warns(UserWarning, match="The frequency is not 'M'."):
                xs.diagnostics.health_checks(ds, freq="M")
        else:
            ds = xr.concat(
                [ds.isel(time=slice(0, 100)), ds.isel(time=slice(200, 365))], dim="time"
            )
            with pytest.raises(
                ValueError,
                match="The timesteps are irregular or cannot be inferred by xarray.",
            ):
                xs.diagnostics.health_checks(ds, freq="D", raise_on=["all"])
            with pytest.warns(
                UserWarning,
                match="The timesteps are irregular or cannot be inferred by xarray.",
            ):
                xs.diagnostics.health_checks(ds, freq="D")
            with pytest.warns(
                UserWarning,
                match="Frequency None is not supported for missing data checks. That check will be skipped.",
            ):
                xs.diagnostics.health_checks(ds, missing="any")

    @pytest.mark.parametrize("missing", ["missing_any", "missing_wmo", "both"])
    def test_missing(self, missing):
        ds = timeseries(
            np.tile(np.arange(1, 366), 3), "tas", "1/1/2001", freq="D", as_dataset=True
        )
        ds = ds.where((ds.time.dt.year > 2001) | (ds.time.dt.dayofyear > 2))

        if missing == "both":
            missing = ["missing_any", "missing_wmo"]

        if missing == "missing_any" or isinstance(missing, list):
            with pytest.raises(
                ValueError,
                match="The variable 'tas' has missing values according to the 'missing_any' method.",
            ):
                xs.diagnostics.health_checks(ds, missing=missing, raise_on=["all"])
            with pytest.warns(
                UserWarning,
                match="The variable 'tas' has missing values according to the 'missing_any' method.",
            ):
                xs.diagnostics.health_checks(ds, missing=missing)
        else:
            xs.diagnostics.health_checks(ds, missing=missing, raise_on=["all"])

    @pytest.mark.parametrize("flag", ["good", "bad"])
    def test_flags(self, flag):
        tasmin = np.array([-15] * 365 * 3)
        tasmax = np.array([15] * 365 * 3)

        if flag == "good":
            flags = {"tasmax": {"tasmax_below_tasmin": {}}}
            ds = timeseries(tasmin, "tasmin", "1/1/2000", freq="D", as_dataset=True)
            ds["tasmax"] = timeseries(tasmax, "tasmax", "1/1/2000", freq="D")
            xs.diagnostics.health_checks(ds, flags=flags, raise_on=["all"])
        else:
            tasmin[7] = 150
            flags = {
                "tasmin": {
                    "temperature_extremely_high": {},
                    "values_repeating_for_n_or_more_days": {"n": 10},
                },
                "tasmax": {
                    "tasmax_below_tasmin": {},
                    "values_repeating_for_n_or_more_days": {"n": 10},
                },
            }
            ds = timeseries(tasmin, "tasmin", "1/1/2000", freq="D", as_dataset=True)
            ds["tasmax"] = timeseries(tasmax, "tasmax", "1/1/2000", freq="D")
            with pytest.raises(
                ValueError,
                match="tasmax_below_tasmin",
            ):
                xs.diagnostics.health_checks(ds, flags=flags, raise_on=["all"])
            with pytest.warns(UserWarning, match="tasmax_below_tasmin"):
                xs.diagnostics.health_checks(ds, flags=flags)
                dsflags = xs.diagnostics.health_checks(
                    ds, flags=flags, flags_kwargs={"freq": "D"}, return_flags=True
                )

                assert len(dsflags.time) == len(ds.time)
                assert len(dsflags.data_vars) == 4
                assert all(
                    [
                        v in dsflags.data_vars
                        for v in [
                            "tasmin_temperature_extremely_high",
                            "tasmin_values_repeating_for_10_or_more_days",
                            "tasmax_tasmax_below_tasmin",
                            "tasmax_values_repeating_for_10_or_more_days",
                        ]
                    ]
                )


class TestPropertiesMeasures:
    yaml_file = notebooks / "samples" / "properties.yml"
    ds = timeseries(
        np.ones(365 * 3), variable="tas", start="2001-01-01", freq="D", as_dataset=True
    )

    @pytest.mark.parametrize("input", ["module", "iter"])
    def test_input_types(self, input):
        module = xs.indicators.load_xclim_module(self.yaml_file)
        p1, m1 = xs.properties_and_measures(
            self.ds,
            properties=module if input == "module" else module.iter_indicators(),
        )
        p2, m2 = xs.properties_and_measures(self.ds, properties=self.yaml_file)
        assert p1.equals(p2)
        assert m1.equals(m2)

    @pytest.mark.parametrize("to_level", [None, "test"])
    def test_level(self, to_level):
        if to_level is None:
            p, m = xs.properties_and_measures(self.ds, properties=self.yaml_file)
            assert "diag-properties" == p.attrs["cat:processing_level"]
            assert "diag-measures" == m.attrs["cat:processing_level"]

        else:
            p, m = xs.properties_and_measures(
                self.ds,
                properties=self.yaml_file,
                to_level_prop=to_level,
                to_level_meas=to_level,
            )
            assert to_level == p.attrs["cat:processing_level"]
            assert to_level == m.attrs["cat:processing_level"]

    @pytest.mark.parametrize("period", [None, ["2001", "2001"]])
    def test_output(self, period):
        values = np.ones(365 * 2)
        values[:365] = 2
        ds = timeseries(
            values, variable="tas", start="2001-01-01", freq="D", as_dataset=True
        )
        ds["da"] = ds.tas

        p, m = xs.properties_and_measures(
            ds,
            properties=self.yaml_file,
            period=period,
        )

        if period is None:
            np.testing.assert_allclose(p["quantile_98_tas"].values, 2)
            np.testing.assert_allclose(p["mean-tas"].values, 1.5)
        else:
            np.testing.assert_allclose(p["quantile_98_tas"].values, 2)
            np.testing.assert_allclose(p["mean-tas"].values, 2)

    def test_unstack(self):
        ds = datablock_3d(
            np.array([[[0, 1, 2], [1, 2, 3], [2, 3, 4]]] * 3, "float"),
            "tas",
            "lon",
            -70,
            "lat",
            15,
            30,
            30,
            as_dataset=True,
        )

        ds_stack = xs.utils.stack_drop_nans(
            ds,
            mask=xr.where(ds.tas.isel(time=0).isnull(), False, True).drop_vars("time"),
        )

        p, m = xs.properties_and_measures(
            ds_stack,
            properties=self.yaml_file,
            unstack=True,
        )

        assert "lat" in p.dims
        assert "lon" in p.dims
        assert "loc" not in p.dims

    def test_rechunk(self):
        ds = datablock_3d(
            np.array([[[0, 1, 2], [1, 2, 3], [2, 3, 4]]] * 3, "float"),
            "tas",
            "lon",
            -70,
            "lat",
            15,
            30,
            30,
            as_dataset=True,
        )
        p, m = xs.properties_and_measures(
            ds,
            properties=self.yaml_file,
            rechunk={"lat": 1, "lon": 1},
        )

        assert p.chunks["lat"] == (1, 1, 1)
        assert p.chunks["lon"] == (1, 1, 1)

    def test_units(self):
        p, m = xs.properties_and_measures(
            self.ds,
            properties=self.yaml_file,
            change_units_arg={"tas": "degC"},
        )

        assert p["mean-tas"].attrs["units"] == "°C"

    def test_dref_for_measure(self):
        p1, m1 = xs.properties_and_measures(
            self.ds,
            properties=self.yaml_file,
        )

        p2, m2 = xs.properties_and_measures(
            self.ds,
            properties=self.yaml_file,
            dref_for_measure=p1,
        )
        print(m2)
        print(m2["maximum_length_of_warm_spell"].values)
        assert m1.sizes == {}
        np.testing.assert_allclose(m2["maximum_length_of_warm_spell"].values, 0)

    def test_measures_heatmap(self):

        p1, m1 = xs.properties_and_measures(
            self.ds,
            properties=self.yaml_file,
        )

        p2, m2 = xs.properties_and_measures(
            self.ds,
            properties=self.yaml_file,
            dref_for_measure=p1,
        )

        out = xs.diagnostics.measures_heatmap({"m2": m2}, to_level="test")

        assert out.attrs["cat:processing_level"] == "test"
        assert "m2" in out.realization.values
        assert "mean-tas" in out.properties.values
        np.testing.assert_allclose(out["heatmap"].values, 0.5)

    def test_measures_improvement(self):

        p1, m1 = xs.properties_and_measures(
            self.ds,
            properties=self.yaml_file,
            period=["2001", "2001"],
        )

        p2, m2 = xs.properties_and_measures(
            self.ds,
            properties=self.yaml_file,
            dref_for_measure=p1,
            period=["2001", "2001"],
        )
        with pytest.warns(
            UserWarning,
            match="meas_datasets has more than 2 datasets."
            + " Only the first 2 will be compared.",
        ):
            out = xs.diagnostics.measures_improvement([m2, m2, m2], to_level="test")

        assert out.attrs["cat:processing_level"] == "test"
        assert "mean-tas" in out.properties.values
        np.testing.assert_allclose(out["improved_grid_points"].values, 1)

    def test_measures_improvement_dim(self):
        p1, m1 = xs.properties_and_measures(
            self.ds,
            properties=self.yaml_file,
            period=["2001", "2001"],
        )

        p2, m2 = xs.properties_and_measures(
            self.ds,
            properties=self.yaml_file,
            dref_for_measure=p1,
            period=["2001", "2001"],
        )
        m3 = xr.concat(
            [
                m2.quantile_98_tas.to_dataset().expand_dims({"dummy_dim": [i]})
                for i in range(3)
            ],
            dim="dummy_dim",
        )
        out = xs.diagnostics.measures_improvement(
            [m3, m3], dim="dummy_dim", to_level="test"
        )
        assert set(out.dims) == {"properties", "season"}
        np.testing.assert_allclose(out["improved_grid_points"].values, 1)

    def test_measures_improvement_2d(self):

        p1, m1 = xs.properties_and_measures(
            self.ds,
            properties=self.yaml_file,
        )

        p2, m2 = xs.properties_and_measures(
            self.ds,
            properties=self.yaml_file,
            dref_for_measure=p1,
        )

        imp = xs.diagnostics.measures_improvement([m2, m2], to_level="test")

        out = xs.diagnostics.measures_improvement_2d(
            {"i1": imp, "i2": imp}, to_level="test"
        )

        assert out.attrs["cat:processing_level"] == "test"
        assert "mean-tas" in out.properties.values
        assert "i1" in out.realization.values
        assert "i2" in out.realization.values
        np.testing.assert_allclose(out["improved_grid_points"].values, 1)

        out2 = xs.diagnostics.measures_improvement_2d(
            {
                "i1": [m2, m2],
                "i2": {"a": m2, "b": m2},
            },
            to_level="test",
        )

        assert out.equals(out2)



"""Input/Output functions for xscen."""

import datetime
import json
import logging
import os
import shutil as sh
from collections import defaultdict
from collections.abc import Sequence
from inspect import signature
from pathlib import Path
from zipfile import ZipFile

import h5py
import netCDF4
import numpy as np
import pandas as pd
import xarray as xr
import zarr
from numcodecs.bitround import BitRound
from rechunker import rechunk as _rechunk
from xclim.core.options import METADATA_LOCALES
from xclim.core.options import OPTIONS as XC_OPTIONS

from .config import parse_config
from .scripting import TimeoutException
from .utils import TRANSLATOR, season_sort_key, strip_cat_attrs, translate_time_chunk

logger = logging.getLogger(__name__)
KEEPBITS = defaultdict(lambda: 12)


__all__ = [
    "clean_incomplete",
    "estimate_chunks",
    "get_engine",
    "make_toc",
    "rechunk",
    "rechunk_for_saving",
    "round_bits",
    "save_to_netcdf",
    "save_to_table",
    "save_to_zarr",
    "subset_maxsize",
    "to_table",
    "unzip_directory",
    "zip_directory",
]


def get_engine(file: str | os.PathLike) -> str:
    """Determine which Xarray engine should be used to open the given file.

    The .zarr, .zarr.zip and .zip extensions are recognized as Zarr datasets,
    the rest is seen as a netCDF. If the file is HDF5, the h5netcdf engine is used.

    Parameters
    ----------
    file : str or os.PathLike
        Path to the file.

    Returns
    -------
    str
        Engine to use with xarray
    """
    # find the ideal engine for xr.open_mfdataset
    if Path(file).suffix in [".zarr", ".zip", ".zarr.zip"]:
        engine = "zarr"
    elif h5py.is_hdf5(file):
        engine = "h5netcdf"
    else:
        engine = "netcdf4"

    return engine


def estimate_chunks(  # noqa: C901
    ds: str | os.PathLike | xr.Dataset,
    dims: list,
    target_mb: float = 50,
    chunk_per_variable: bool = False,
) -> dict:
    """Return an approximate chunking for a file or dataset.

    Parameters
    ----------
    ds : xr.Dataset, str
        Either a xr.Dataset or the path to a NetCDF file. Existing chunks are not taken into account.
    dims : list
        Dimension(s) on which to estimate the chunking. Not implemented for more than 2 dimensions.
    target_mb : float
        Roughly the size of chunks (in Mb) to aim for.
    chunk_per_variable : bool
        If True, the output will be separated per variable. Otherwise, a common chunking will be found.

    Returns
    -------
    dict
        A dictionary mapping dimensions to chunk sizes.
    """

    def _estimate_chunks(da, target_mb, size_of_slice, rechunk_dims):
        # Divide the dimensions by the smallest dimension
        min_dim = np.min([da[d].shape[0] for d in rechunk_dims])
        ratio = {d: da[d].shape[0] / min_dim for d in rechunk_dims}

        # Get the approximate number of chunks, supposing the chunks are cubes
        approx_chunks = np.power(target_mb / size_of_slice, 1 / len(rechunk_dims))

        # Redistribute the chunks based on the ratio of the dimensions
        x = (approx_chunks ** len(rechunk_dims) / np.prod(list(ratio.values()))) ** (
            1 / len(rechunk_dims)
        )
        rounding_per_dim = {
            d: 1 if da[d].shape[0] <= 15 else 5 if da[d].shape[0] <= 250 else 10
            for d in rechunk_dims
        }
        chunks_per_dim = {
            d: int(rounding_per_dim[d] * np.round(x * ratio[d] / rounding_per_dim[d]))
            for d in rechunk_dims
        }
        chunks_per_dim = {
            d: np.max([np.min([chunks_per_dim[d], da[d].shape[0]]), 1])
            for d in rechunk_dims
        }

        return chunks_per_dim

    out = {}
    # If ds is the path to a file, use NetCDF4
    if isinstance(ds, str | os.PathLike):
        ds = netCDF4.Dataset(ds, "r")

        # Loop on variables
        for v in ds.variables:
            # Find dimensions to chunk
            rechunk_dims = list(set(dims).intersection(ds.variables[v].dimensions))
            if not rechunk_dims or v in ds.dimensions:
                continue

            dtype_size = ds.variables[v].datatype.itemsize
            num_elem_per_slice = np.prod(
                [ds[d].shape[0] for d in ds[v].dimensions if d not in rechunk_dims]
            )

            size_of_slice = (num_elem_per_slice * dtype_size) / 1024**2

            estimated_chunks = _estimate_chunks(
                ds, target_mb, size_of_slice, rechunk_dims
            )
            for other in set(ds[v].dimensions).difference(dims):
                estimated_chunks[other] = -1

            if chunk_per_variable:
                out[v] = estimated_chunks
            else:
                for d in estimated_chunks:
                    if (d not in out) or (out[d] > estimated_chunks[d]):
                        out[d] = estimated_chunks[d]

    # Else, use xarray
    else:
        for v in ds.data_vars:
            # Find dimensions to chunk
            rechunk_dims = list(set(dims).intersection(ds[v].dims))
            if not rechunk_dims:
                continue

            dtype_size = ds[v].dtype.itemsize
            num_elem_per_slice = np.prod(
                [ds[d].shape[0] for d in ds[v].dims if d not in rechunk_dims]
            )
            size_of_slice = (num_elem_per_slice * dtype_size) / 1024**2

            estimated_chunks = _estimate_chunks(
                ds, target_mb, size_of_slice, rechunk_dims
            )
            for other in set(ds[v].dims).difference(dims):
                estimated_chunks[other] = -1

            if chunk_per_variable:
                out[v] = estimated_chunks
            else:
                for d in estimated_chunks:
                    if (d not in out) or (out[d] > estimated_chunks[d]):
                        out[d] = estimated_chunks[d]

    return out


def subset_maxsize(
    ds: xr.Dataset,
    maxsize_gb: float,
) -> list[xr.Dataset]:
    """Estimate a dataset's size and, if higher than the given limit, subset it alongside the 'time' dimension.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to be saved.
    maxsize_gb : float
        Target size for the NetCDF files.
        If the dataset is bigger than this number, it will be separated alongside the 'time' dimension.

    Returns
    -------
    list of xr.Dataset
        List of xr.Dataset subsetted alongside 'time' to limit the filesize to the requested maximum.
    """
    # Estimate the size of the dataset
    size_of_file = 0
    for v in ds:
        dtype_size = ds[v].dtype.itemsize
        varsize = np.prod(list(ds[v].sizes.values()))
        size_of_file = size_of_file + (varsize * dtype_size) / 1024**3

    if size_of_file < maxsize_gb:
        msg = f"Dataset is already smaller than {maxsize_gb} Gb."
        logger.info(msg)
        return [ds]

    elif "time" in ds.dims:
        years = np.unique(ds.time.dt.year)
        ratio = np.max([int(len(years) / (size_of_file / maxsize_gb)), 1])
        ds_sub = []
        for y in range(years[0], years[-1] + 1, ratio):
            ds_sub.extend([ds.sel({"time": slice(str(y), str(y + ratio - 1))})])
        return ds_sub

    else:
        raise NotImplementedError(
            f"Size of the NetCDF file exceeds the {maxsize_gb} Gb target, but the dataset does not contain a 'time' variable."
        )


def clean_incomplete(
    path: str | os.PathLike,
    complete: Sequence[str] | None = None,
    incomplete: Sequence[str] | None = None,
) -> None:
    """Delete un-catalogued variables from a zarr folder.

    The goal of this function is to clean up an incomplete calculation.
    It will remove any variable in the zarr that is neither in the `complete` list
    nor in the  `coords`.

    Parameters
    ----------
    path : str, Path
        A path to a zarr folder.
    complete : sequence of strings, optional
        Name of variables that were completed. All other variables (except coordinates) will be removed.
        Use either `complete` or `incomplete`, not both.
    incomplete : sequence of strings, optional
        Name of variables that should be removed. Coordinates and dimensions will never be removed through this function.
        Use either `complete` or `incomplete`, not both.

    Returns
    -------
    None
    """
    path = Path(path)

    if complete is not None and incomplete is not None:
        raise ValueError("Use either `complete` or `incomplete`, not both.")
    v_to_rm = []

    if complete is not None:
        with xr.open_zarr(path) as ds:
            complete = set(complete).union(ds.coords.keys())

        for fold in filter(lambda p: p.is_dir(), path.iterdir()):
            if fold.name not in complete:
                v_to_rm.append(fold.name)
                msg = f"Removing {fold} from disk"
                logger.warning(msg)
                sh.rmtree(fold)

    elif incomplete is not None:
        with xr.open_zarr(path) as ds:
            incomplete = [
                v for v in incomplete if (v not in ds.coords) and (v not in ds.dims)
            ]

        for fold in filter(lambda p: p.is_dir(), path.iterdir()):
            if fold.name in incomplete:
                v_to_rm.append(fold.name)
                msg = f"Removing {fold} from disk"
                logger.warning(msg)
                sh.rmtree(fold)

    # Update the .zmetadata file
    with (path / ".zmetadata").open("r") as f:
        metadata = json.load(f)
    for v in v_to_rm:
        [
            metadata["metadata"].pop(k)
            for k in list(metadata["metadata"].keys())
            if k.startswith(f"{v}/.")
        ]
    with (path / ".zmetadata").open("w") as f:
        json.dump(metadata, f, indent=2)


def _coerce_attrs(attrs):
    """Ensure no funky objects in attrs."""
    for k in list(attrs.keys()):
        if not (
            isinstance(attrs[k], str | float | int | np.ndarray)
            or isinstance(attrs[k], tuple | list)
            and isinstance(attrs[k][0], str | float | int)
        ):
            attrs[k] = str(attrs[k])


def _np_bitround(array: xr.DataArray, keepbits: int):
    """Bitround for Arrays."""
    codec = BitRound(keepbits=keepbits)
    data = array.copy()  # otherwise overwrites the input
    encoded = codec.encode(data)
    return codec.decode(encoded)


def round_bits(da: xr.DataArray, keepbits: int):
    """Round floating point variable by keeping a given number of bits in the mantissa, dropping the rest. This allows for a much better compression.

    Parameters
    ----------
    da : xr.DataArray
        Variable to be rounded.
    keepbits : int
        The number of bits of the mantissa to keep.
    """
    encoding = da.encoding
    da = xr.apply_ufunc(
        _np_bitround, da, keepbits, dask="parallelized", keep_attrs=True
    )
    da.encoding = encoding
    da.attrs["_QuantizeBitRoundNumberOfSignificantDigits"] = keepbits
    new_history = f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] Data compressed with BitRound by keeping {keepbits} bits."
    history = (
        new_history + " \n " + da.attrs["history"]
        if "history" in da.attrs
        else new_history
    )
    da.attrs["history"] = history
    return da


def _get_keepbits(bitround: bool | int | dict, varname: str, vartype):
    # Guess the number of bits to keep depending on how bitround was passed, the var dtype and the var name.
    if not np.issubdtype(vartype, np.floating) or bitround is False:
        if isinstance(bitround, dict) and varname in bitround:
            raise ValueError(
                f"A keepbits value was given for variable {varname} even though it is not of a floating dtype."
            )
        return None
    if bitround is True:
        return KEEPBITS[varname]
    if isinstance(bitround, int):
        return bitround
    if isinstance(bitround, dict):
        return bitround.get(varname, KEEPBITS[varname])
    return None


@parse_config
def save_to_netcdf(
    ds: xr.Dataset,
    filename: str | os.PathLike,
    *,
    rechunk: dict | None = None,
    bitround: bool | int | dict = False,
    compute: bool = True,
    netcdf_kwargs: dict | None = None,
    strip_cat_metadata: bool = True,
):
    """Save a Dataset to NetCDF, rechunking or compressing if requested.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to be saved.
    filename : str or os.PathLike
        Name of the NetCDF file to be saved.
    rechunk : dict, optional
        This is a mapping from dimension name to new chunks (in any format understood by dask).
        Spatial dimensions can be generalized as 'X' and 'Y', which will be mapped to the actual grid type's
        dimension names.
        Rechunking is only done on *data* variables sharing dimensions with this argument.
    bitround : bool or int or dict
        If not False, float variables are bit-rounded by dropping a certain number of bits from their mantissa,
        allowing for a much better compression.
        If an int, this is the number of bits to keep for all float variables.
        If a dict, a mapping from variable name to the number of bits to keep.
        If True, the number of bits to keep is guessed based on the variable's name, defaulting to 12,
        which yields a relative error below 0.013%.
    compute : bool
        Whether to start the computation or return a delayed object.
    netcdf_kwargs : dict, optional
        Additional arguments to send to_netcdf()
    strip_cat_metadata : bool
        If True (default), strips all catalog-added attributes before saving the dataset.

    Returns
    -------
    None

    See Also
    --------
    xarray.Dataset.to_netcdf
    """
    if rechunk:
        ds = rechunk_for_saving(ds, rechunk)

    path = Path(filename)
    path.parent.mkdir(parents=True, exist_ok=True)

    # Prepare to_netcdf kwargs
    netcdf_kwargs = netcdf_kwargs or {}
    netcdf_kwargs.setdefault("engine", "h5netcdf")
    netcdf_kwargs.setdefault("format", "NETCDF4")

    for var in list(ds.data_vars.keys()):
        if keepbits := _get_keepbits(bitround, var, ds[var].dtype):
            ds = ds.assign({var: round_bits(ds[var], keepbits)})
        # Remove original_shape from encoding, since it can cause issues with some engines.
        ds[var].encoding.pop("original_shape", None)

    if strip_cat_metadata:
        ds = strip_cat_attrs(ds)

    _coerce_attrs(ds.attrs)
    for var in ds.variables.values():
        _coerce_attrs(var.attrs)

    return ds.to_netcdf(filename, compute=compute, **netcdf_kwargs)


@parse_config
def save_to_zarr(  # noqa: C901
    ds: xr.Dataset,
    filename: str | os.PathLike,
    *,
    rechunk: dict | None = None,
    zarr_kwargs: dict | None = None,
    compute: bool = True,
    encoding: dict | None = None,
    bitround: bool | int | dict = False,
    mode: str = "f",
    itervar: bool = False,
    timeout_cleanup: bool = True,
    strip_cat_metadata: bool = True,
):
    """
    Save a Dataset to Zarr format, rechunking and compressing if requested.

    According to mode, removes variables that we don't want to re-compute in ds.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to be saved.
    filename : str
        Name of the Zarr file to be saved.
    rechunk : dict, optional
        This is a mapping from dimension name to new chunks (in any format understood by dask).
        Spatial dimensions can be generalized as 'X' and 'Y' which will be mapped to the actual grid type's
        dimension names.
        Rechunking is only done on *data* variables sharing dimensions with this argument.
    zarr_kwargs : dict, optional
        Additional arguments to send to_zarr()
    compute : bool
        Whether to start the computation or return a delayed object.
    mode : {'f', 'o', 'a'}
        If 'f', fails if any variable already exists.
        if 'o', removes the existing variables.
        if 'a', skip existing variables, writes the others.
    encoding : dict, optional
        If given here instead of 'zarr_kwargs', encoding will only be applied to the variables that are being written,
        skipping those that are already in the zarr.
    bitround : bool or int or dict
        If not False, float variables are bit-rounded by dropping a certain number of bits from their mantissa,
        allowing for a much better compression.
        If an int, this is the number of bits to keep for all float variables.
        If a dict, a mapping from variable name to the number of bits to keep.
        If True, the number of bits to keep is guessed based on the variable's name, defaulting to 12,
        which yields a relative error of 0.012%.
    itervar : bool
        If True, (data) variables are written one at a time, appending to the zarr.
        If False, this function computes, no matter what was passed to kwargs.
    timeout_cleanup : bool
        If True (default) and a :py:class:`xscen.scripting.TimeoutException` is raised during the writing,
        the variable being written is removed from the dataset as it is incomplete.
        This does nothing if `compute` is False.
    strip_cat_metadata : bool
        If True (default), strips all catalog-added attributes before saving the dataset.

    Returns
    -------
    dask.delayed object if compute=False, None otherwise.

    See Also
    --------
    xarray.Dataset.to_zarr
    """
    # to address this issue https://github.com/pydata/xarray/issues/3476
    for v in list(ds.coords.keys()):
        if ds.coords[v].dtype == object:
            ds[v].encoding.clear()

    if rechunk:
        ds = rechunk_for_saving(ds, rechunk)

    path = Path(filename)
    path.parent.mkdir(parents=True, exist_ok=True)
    if path.is_dir():
        tgtds = zarr.open(str(path), mode="r")
    else:
        tgtds = {}

    if encoding:
        encoding = encoding.copy()

    # Prepare to_zarr kwargs
    if zarr_kwargs is None:
        zarr_kwargs = {}

    def _skip(var):
        exists = var in tgtds

        if mode == "f" and exists:
            raise ValueError(f"Variable {var} exists in dataset {path}.")

        if mode == "o":
            if exists:
                clean_incomplete(path, incomplete=[var])
            return False

        if mode == "a":
            # In all cases, we need to skip the encoding of existing variables.
            if exists:
                if encoding:
                    encoding.pop(var, None)

            # If we are not appending, we need to skip the writing of existing variables.
            if "append_dim" not in zarr_kwargs:
                return exists

            # If we are appending, we need to raise an error if there are new variables.
            elif exists is False:
                raise ValueError(
                    f"When 'append_dim' is set in zarr_kwargs, all variables must already exist in the dataset."
                )

            return False

    for var in list(ds.data_vars.keys()):
        if _skip(var):
            msg = f"Skipping {var} in {path}."
            logger.info(msg)
            ds = ds.drop_vars(var)
            continue
        if keepbits := _get_keepbits(bitround, var, ds[var].dtype):
            ds = ds.assign({var: round_bits(ds[var], keepbits)})
        # Remove original_shape from encoding, since it can cause issues with some engines.
        ds[var].encoding.pop("original_shape", None)

    if len(ds.data_vars) == 0:
        return None

    if strip_cat_metadata:
        ds = strip_cat_attrs(ds)

    _coerce_attrs(ds.attrs)
    for var in ds.variables.values():
        _coerce_attrs(var.attrs)

    if itervar:
        zarr_kwargs["compute"] = True
        allvars = set(ds.data_vars.keys())
        if mode == "f":
            dsbase = ds.drop_vars(allvars)
            dsbase.to_zarr(path, **zarr_kwargs)
        if mode == "o":
            dsbase = ds.drop_vars(allvars)
            dsbase.to_zarr(path, **zarr_kwargs, mode="w")
        for i, (name, var) in enumerate(ds.data_vars.items()):
            msg = f"Writing {name} ({i + 1} of {len(ds.data_vars)}) to {path}"
            logger.debug(msg)
            dsvar = ds.drop_vars(allvars - {name})
            try:
                dsvar.to_zarr(
                    path,
                    mode="a",
                    encoding={k: v for k, v in (encoding or {}).items() if k in dsvar},
                    **zarr_kwargs,
                )
            except TimeoutException:
                if timeout_cleanup:
                    clean_incomplete(path, incomplete=[name])
                raise

    else:
        msg = f"Writing {list(ds.data_vars.keys())} for {filename}."
        logger.debug(msg)
        try:
            return ds.to_zarr(
                filename, compute=compute, mode="a", encoding=encoding, **zarr_kwargs
            )
        except TimeoutException:
            if timeout_cleanup:
                clean_incomplete(path, incomplete=list(ds.data_vars.keys()))
            raise


def _to_dataframe(
    data: xr.DataArray,
    row: list[str],
    column: list[str],
    coords: list[str],
    coords_dims: dict,
):
    """Convert a DataArray to a DataFrame with support for MultiColumn."""
    df = data.to_dataframe()
    if not column:
        # Fast track for the easy case where xarray's default is already what we want.
        return df
    df_data = (
        df[[data.name]]
        .reset_index()
        .pivot(index=row, columns=column)
        .droplevel(None, axis=1)
    )
    dfs = []
    for v in coords:
        drop_cols = [c for c in column if c not in coords_dims[v]]
        cols = [c for c in column if c in coords_dims[v]]
        dfc = (
            df[[v]].reset_index().drop(columns=drop_cols).pivot(index=row, columns=cols)
        )
        cols = dfc.columns
        # The "None" level has the aux coord name we want it either at the same level as variable, or at lowest missing level otherwise.
        varname_lvl = "variable" if "variable" in drop_cols else drop_cols[-1]
        cols = cols.rename(
            varname_lvl
            if not isinstance(cols, pd.MultiIndex)
            else [nm or varname_lvl for nm in cols.name]
        )
        if isinstance(df_data.columns, pd.MultiIndex) or isinstance(
            cols, pd.MultiIndex
        ):
            # handle different depth of multicolumns, expand MultiCol of coord with None for missing levels.
            cols = pd.MultiIndex.from_arrays(
                [
                    cols.get_level_values(lvl) if lvl in cols.names else [None]
                    for lvl in df_data.columns.names
                ],
                names=df_data.columns.names,
            )
        dfc.columns = cols
        dfs.append(
            dfc[~dfc.index.duplicated()]
        )  # We dropped columns thus the index is not unique anymore
    dfs.append(df_data)
    return pd.concat(dfs, axis=1).sort_index(level=row, key=season_sort_key)


def to_table(
    ds: xr.Dataset | xr.DataArray,
    *,
    row: str | Sequence[str] | None = None,
    column: str | Sequence[str] | None = None,
    sheet: str | Sequence[str] | None = None,
    coords: bool | str | Sequence[str] = True,
) -> pd.DataFrame | dict:
    """Convert a dataset to a pandas DataFrame with support for multicolumns and multisheet.

    This function will trigger a computation of the dataset.

    Parameters
    ----------
    ds : xr.Dataset or xr.DataArray
      Dataset or DataArray to be saved.
      If a Dataset with more than one variable is given, the dimension "variable"
      must appear in one of `row`, `column` or `sheet`.
    row : str or sequence of str, optional
      Name of the dimension(s) to use as indexes (rows).
      Default is all data dimensions.
    column : str or sequence of str, optional
      Name of the dimension(s) to use as columns.
      Default is "variable", i.e. the name of the variable(s).
    sheet : str or sequence of str, optional
      Name of the dimension(s) to use as sheet names.
    coords: bool or str or sequence of str
      A list of auxiliary coordinates to add to the columns (as would variables).
      If True, all (if any) are added.

    Returns
    -------
    pd.DataFrame or dict
      DataFrame with a MultiIndex with levels `row` and MultiColumn with levels `column`.
      If `sheet` is given, the output is dictionary with keys for each unique "sheet" dimensions tuple, values are DataFrames.
      The DataFrames are always sorted with level priority as given in `row` and in ascending order.
    """
    if isinstance(ds, xr.Dataset):
        da = ds.to_array(name="data")
        if len(ds) == 1:
            da = da.isel(variable=0)
            da.name = str(da["variable"].values)
            da = da.drop_vars("variable")
    else:
        da = ds

    def _ensure_list(seq):
        if isinstance(seq, str):
            return [seq]
        return list(seq)

    passed_dims = set().union(
        _ensure_list(row or []), _ensure_list(column or []), _ensure_list(sheet or [])
    )
    if row is None:
        row = [d for d in da.dims if d != "variable" and d not in passed_dims]
    row = _ensure_list(row)
    if column is None:
        column = (
            ["variable"]
            if isinstance(ds, xr.Dataset)
            and len(ds) > 1
            and "variable" not in passed_dims
            else []
        )
    column = _ensure_list(column)
    if sheet is None:
        sheet = []
    sheet = _ensure_list(sheet)

    needed_dims = row + column + sheet
    if len(set(needed_dims)) != len(needed_dims):
        raise ValueError(
            f"Repeated dimension names. Got row={row}, column={column} and sheet={sheet}."
            "Each dimension should appear only once."
        )
    if set(needed_dims) != set(da.dims):
        raise ValueError(
            f"Passed row, column and sheet do not match available dimensions. Got {needed_dims}, data has {da.dims}."
        )

    if coords is not True:
        coords = _ensure_list(coords or [])
        drop = set(da.coords.keys()) - set(da.dims) - set(coords)
        da = da.drop_vars(drop)
    else:
        coords = list(set(da.coords.keys()) - set(da.dims))
    if len(coords) > 1 and ("variable" in row or "variable" in sheet):
        raise NotImplementedError(
            "Keeping auxiliary coords is not implemented when 'variable' is in the row or in the sheets."
            "Pass `coords=False` or put 'variable' in `column` instead."
        )

    table_kwargs = dict(
        row=row,
        column=column,
        coords=coords,
        coords_dims={c: ds[c].dims for c in coords},
    )
    if sheet:
        out = {}
        das = da.stack(sheet=sheet)
        for elem in das.sheet:
            out[elem.item()] = _to_dataframe(
                das.sel(sheet=elem, drop=True), **table_kwargs
            )
        return out
    return _to_dataframe(da, **table_kwargs)


def make_toc(ds: xr.Dataset | xr.DataArray, loc: str | None = None) -> pd.DataFrame:
    """Make a table of content describing a dataset's variables.

    This return a simple DataFrame with variable names as index, the long_name as "description" and units.
    Column names and long names are taken from the activated locale if found, otherwise the english version is taken.

    Parameters
    ----------
    ds : xr.Dataset or xr.DataArray
      Dataset or DataArray from which to extract the relevant metadata.
    loc : str, optional
        The locale to use. If None, either the first locale in the list of activated xclim locales is used, or "en" if none is activated.

    Returns
    -------
    pd.DataFrame
      A DataFrame with variables as index, and columns "description" and "units".
    """
    if loc is None:
        loc = (XC_OPTIONS[METADATA_LOCALES] or ["en"])[0]
    locsuf = "" if loc == "en" else f"_{loc}"
    _ = TRANSLATOR[loc]  # Combine translation and gettext parsing (like it usually is)

    if isinstance(ds, xr.DataArray):
        ds = ds.to_dataset()

    toc = pd.DataFrame.from_records(
        [
            {
                _("Variable"): vv,
                _("Description"): da.attrs.get(
                    f"long_name{locsuf}", da.attrs.get("long_name")
                ),
                _("Units"): da.attrs.get("units"),
            }
            for vv, da in ds.data_vars.items()
        ],
    ).set_index(_("Variable"))

    # Add global attributes by using a fake variable and description
    if len(ds.attrs) > 0:
        globattr = pd.DataFrame.from_records(
            [
                {
                    _("Variable"): vv,
                    _("Description"): da,
                    _("Units"): "",
                }
                for vv, da in ds.attrs.items()
            ],
        ).set_index(_("Variable"))
        globattr.attrs["name"] = _("Global attributes")

        # Empty row to separate global attributes from variables
        toc = pd.concat([toc, pd.DataFrame(index=[""])])
        toc = pd.concat([toc, pd.DataFrame(index=[_("Global attributes")])])
        toc = pd.concat([toc, globattr])
    toc.attrs["name"] = _("Content")

    return toc


TABLE_FORMATS = {".csv": "csv", ".xls": "excel", ".xlsx": "excel"}


def save_to_table(  # noqa: C901
    ds: xr.Dataset | xr.DataArray,
    filename: str | os.PathLike,
    output_format: str | None = None,
    *,
    row: str | Sequence[str] | None = None,
    column: None | str | Sequence[str] = None,
    sheet: str | Sequence[str] | None = None,
    coords: bool | Sequence[str] = True,
    col_sep: str = "_",
    row_sep: str | None = None,
    add_toc: bool | pd.DataFrame = False,
    **kwargs,
):
    r"""Save the dataset to a tabular file (csv, excel, ...).

    This function will trigger a computation of the dataset.

    Parameters
    ----------
    ds : xr.Dataset or xr.DataArray
      Dataset or DataArray to be saved.
      If a Dataset with more than one variable is given, the dimension "variable"
      must appear in one of `row`, `column` or `sheet`.
    filename : str or os.PathLike
      Name of the file to be saved.
    output_format: {'csv', 'excel', ...}, optional
      The output format. If None (default), it is inferred
      from the extension of `filename`. Not all possible output format are supported for inference.
      Valid values are any that matches a :py:class:`pandas.DataFrame` method like "df.to_{format}".
    row : str or sequence of str, optional
      Name of the dimension(s) to use as indexes (rows).
      Default is all data dimensions.
    column : str or sequence of str, optional
      Name of the dimension(s) to use as columns.
      When using a Dataset with more than 1 variable, default is "variable", i.e. the name of the variable(s).
      When using a DataArray, default is None.
    sheet : str or sequence of str, optional
      Name of the dimension(s) to use as sheet names.
      Only valid if the output format is excel.
    coords: bool or sequence of str
      A list of auxiliary coordinates to add to the columns (as would variables).
      If True, all (if any) are added.
    col_sep : str,
      Multi-columns (except in excel) and sheet names are concatenated with this separator.
    row_sep : str, optional
      Multi-index names are concatenated with this separator, except in excel.
      If None (default), each level is written in its own column.
    add_toc : bool or DataFrame
      A table of content to add as the first sheet. Only valid if the output format is excel.
      If True, :py:func:`make_toc` is used to generate the toc.
      The sheet name of the toc can be given through the "name" attribute of the DataFrame, otherwise "Content" is used.
    \*\*kwargs:
      Other arguments passed to the pandas function.
      If the output format is excel, kwargs to :py:class:`pandas.ExcelWriter` can be given here as well.
    """
    filename = Path(filename)

    if output_format is None:
        output_format = TABLE_FORMATS.get(filename.suffix)
    if output_format is None:
        raise ValueError(
            f"Output format could not be inferred from filename {filename.name}. Please pass `output_format`."
        )

    if column is None and isinstance(ds, xr.Dataset) and len(ds.data_vars) > 1:
        column = "variable"

    if sheet is not None and output_format != "excel":
        raise ValueError(
            f"Argument `sheet` is only valid with excel as the output format. Got {output_format}."
        )
    if add_toc is not False and output_format != "excel":
        raise ValueError(
            f"A TOC was requested, but the output format is not Excel. Got {output_format}."
        )

    out = to_table(ds, row=row, column=column, sheet=sheet, coords=coords)

    if add_toc is not False:
        if not sheet:
            out = {("data",): out}
        if add_toc is True:
            add_toc = make_toc(ds)
        out = {(add_toc.attrs.get("name", "Content"),): add_toc, **out}

    # Get engine_kwargs
    if output_format == "excel":
        engine_kwargs = {}  # Extract engine kwargs
        for arg in signature(pd.ExcelWriter).parameters:
            if arg in kwargs:
                engine_kwargs[arg] = kwargs.pop(arg)
    else:
        engine_kwargs = {}

    if sheet or (add_toc is not False):
        with pd.ExcelWriter(filename, **engine_kwargs) as writer:
            for sheet_name, df in out.items():
                df.to_excel(writer, sheet_name=col_sep.join(sheet_name), **kwargs)
    elif len(engine_kwargs) > 0:
        with pd.ExcelWriter(filename, **engine_kwargs) as writer:
            out.to_excel(writer, **kwargs)
    else:
        if output_format != "excel" and isinstance(out.columns, pd.MultiIndex):
            out.columns = out.columns.map(lambda lvls: col_sep.join(map(str, lvls)))
        if (
            output_format != "excel"
            and row_sep is not None
            and isinstance(out.index, pd.MultiIndex)
        ):
            new_name = row_sep.join(out.index.names)
            out.index = out.index.map(lambda lvls: row_sep.join(map(str, lvls)))
            out.index.name = new_name
        getattr(out, f"to_{output_format}")(filename, **kwargs)


def rechunk_for_saving(ds: xr.Dataset, rechunk: dict):
    """Rechunk before saving to .zarr or .nc, generalized as Y/X for different axes lat/lon, rlat/rlon.

    Parameters
    ----------
    ds : xr.Dataset
        The xr.Dataset to be rechunked.
    rechunk : dict
        A dictionary with the dimension names of ds and the new chunk size. Spatial dimensions
        can be provided as X/Y.

    Returns
    -------
    xr.Dataset
        The dataset with new chunking.
    """
    for rechunk_var in ds.data_vars:
        # Support for chunks varying per variable
        if rechunk_var in rechunk:
            rechunk_dims = rechunk[rechunk_var].copy()
        else:
            rechunk_dims = rechunk.copy()

        # get actual axes labels
        if "X" in rechunk_dims and "X" not in ds.dims:
            rechunk_dims[ds.cf.axes["X"][0]] = rechunk_dims.pop("X")
        if "Y" in rechunk_dims and "Y" not in ds.dims:
            rechunk_dims[ds.cf.axes["Y"][0]] = rechunk_dims.pop("Y")

        ds[rechunk_var] = ds[rechunk_var].chunk(
            {d: chnks for d, chnks in rechunk_dims.items() if d in ds[rechunk_var].dims}
        )
        ds[rechunk_var].encoding["chunksizes"] = tuple(
            rechunk_dims[d] if d in rechunk_dims else ds[d].shape[0]
            for d in ds[rechunk_var].dims
        )
        ds[rechunk_var].encoding.pop("chunks", None)
        ds[rechunk_var].encoding["preferred_chunks"] = rechunk_dims

    return ds


@parse_config
def rechunk(
    path_in: os.PathLike | str | xr.Dataset,
    path_out: os.PathLike | str,
    *,
    chunks_over_var: dict | None = None,
    chunks_over_dim: dict | None = None,
    worker_mem: str,
    temp_store: os.PathLike | str | None = None,
    overwrite: bool = False,
) -> None:
    """Rechunk a dataset into a new zarr.

    Parameters
    ----------
    path_in : path, str or xr.Dataset
        Input to rechunk.
    path_out : path or str
        Path to the target zarr.
    chunks_over_var : dict
        Mapping from variables to mappings from dimension name to size. Give this argument or `chunks_over_dim`.
    chunks_over_dim : dict
        Mapping from dimension name to size that will be used for all variables in ds.
        Give this argument or `chunks_over_var`.
    worker_mem : str
        The maximal memory usage of each task.
        When using a distributed Client, this an approximate memory per thread.
        Each worker of the client should have access to 10-20% more memory than this times the number of threads.
    temp_store : path or str, optional
        A path to a zarr where to store intermediate results.
    overwrite : bool
        If True, it will delete whatever is in path_out before doing the rechunking.

    Returns
    -------
    None

    See Also
    --------
    rechunker.rechunk
    """
    if Path(path_out).is_dir() and overwrite:
        sh.rmtree(path_out)

    if isinstance(path_in, os.PathLike) or isinstance(path_in, str):
        path_in = Path(path_in)
        if path_in.suffix == ".zarr":
            ds = xr.open_zarr(path_in)
        else:
            ds = xr.open_dataset(path_in)
    else:
        ds = path_in
    # Remove all input chunks information, avoids an error with rechunker and xarray >= 2024.10
    # TODO: Remove this and pin rechunker when https://github.com/pangeo-data/rechunker/pull/156 is merged and released
    for var in ds.variables.values():
        var.encoding.pop("chunks", None)

    variables = list(ds.data_vars)
    if chunks_over_var:
        chunks = chunks_over_var
    elif chunks_over_dim:
        chunks = {v: {d: chunks_over_dim[d] for d in ds[v].dims} for v in variables}
        chunks.update(time=None, lat=None, lon=None)
        cal = ds.time.dt.calendar
        Nt = ds.time.size
        chunks = translate_time_chunk(chunks, cal, Nt)
    else:
        raise ValueError(
            "No chunks given. Need to give at `chunks_over_var` or `chunks_over_dim`."
        )
    plan = _rechunk(ds, chunks, worker_mem, str(path_out), temp_store=str(temp_store))

    plan.execute()
    zarr.consolidate_metadata(path_out)

    if temp_store is not None:
        sh.rmtree(temp_store)


def zip_directory(
    root: str | os.PathLike,
    zipfile: str | os.PathLike,
    delete: bool = False,
    **zip_args,
):
    r"""Make a zip archive of the content of a directory.

    Parameters
    ----------
    root : path
        The directory with the content to archive.
    zipfile : path
        The zip file to create.
    delete : bool
        If True, the original directory is deleted after zipping.
    \*\*zip_args
        Any other arguments to pass to :py:mod:`zipfile.ZipFile`, such as "compression".
        The default is to make no compression (``compression=ZIP_STORED``).
    """
    root = Path(root)

    def _add_to_zip(zf, path, root):
        zf.write(path, path.relative_to(root))
        if path.is_dir():
            for subpath in path.iterdir():
                _add_to_zip(zf, subpath, root)

    with ZipFile(zipfile, "w", **zip_args) as zf:
        for file in root.iterdir():
            _add_to_zip(zf, file, root)

    if delete:
        sh.rmtree(root)


def unzip_directory(zipfile: str | os.PathLike, root: str | os.PathLike):
    r"""Unzip an archive to a directory.

    This function is the exact opposite of :py:func:`xscen.io.zip_directory`.

    Parameters
    ----------
    zipfile : path
        The zip file to read.
    root : path
        The directory where to put the content to archive.
        If doesn't exist, it will be created (and all its parents).
        If it exists, should be empty.
    """
    root = Path(root)
    root.mkdir(parents=True, exist_ok=True)

    with ZipFile(zipfile, "r") as zf:
        zf.extractall(root)



from collections.abc import Generator
from pathlib import Path

import pandas as pd
import pytest
import xarray as xr
from conftest import SAMPLES_DIR

from xscen import catutils as cu


# Sample files are in the doc folder
@pytest.mark.requires_netcdf
@pytest.mark.parametrize(
    "exts,lens,dirglob,N",
    (
        [{".nc"}, {7}, None, 10],
        [{".zarr"}, {7}, None, 0],
        [{".nc"}, {6}, None, 0],
        [{".nc", ".zarr"}, {6, 7, 8}, "*ssp126*", 2],
    ),
)
def test_find_assets(exts, lens, dirglob, N):  # noqa: N803
    finder = cu._find_assets(str(SAMPLES_DIR), exts=exts, lengths=lens, dirglob=dirglob)
    assert isinstance(finder, Generator)
    assert len(list(finder)) == N


def test_name_parser():
    d = cu._name_parser(
        "station-obs/GovCan/AHCCD/day/tasmin/tasmin_day_GovCan_AHCCD_1960-1999.nc",
        root="",
        patterns=[
            "{type}/{institution}/{source}/{frequency}/{variable}/{?var}_{?:_}_{DATES}.nc"
        ],
    )
    assert d["type"] == "station-obs"
    assert d["date_start"] == "1960"
    assert d["format"] == "nc"
    assert set(d.keys()) == {
        "type",
        "institution",
        "source",
        "frequency",
        "variable",
        "date_start",
        "date_end",
        "path",
        "format",
    }


def test_parse_dir_filechecks(tmp_path):
    file = tmp_path / "this_is_not_a_netcdf_file.nc"
    file.touch()
    file.chmod(0o111)  # remove read and write perms for all

    # No checks
    parsed = cu._parse_dir(tmp_path, patterns=["{?:_}.nc"], checks=None)
    assert len(parsed) == 1

    # Not readable
    parsed = cu._parse_dir(tmp_path, patterns=["{?:_}.nc"], checks=["readable"])
    assert len(parsed) == 0

    # Not writable
    file.chmod(0o444)  # read-only
    parsed = cu._parse_dir(tmp_path, patterns=["{?:_}.nc"], checks=["writable"])
    assert len(parsed) == 0

    # No checks
    file.chmod(0o777)
    parsed = cu._parse_dir(tmp_path, patterns=["{?:_}.nc"], checks=["ncvalid"])
    assert len(parsed) == 0

    # Just to be sure
    parsed = cu._parse_dir(
        tmp_path, patterns=["{?:_}.nc"], checks=["readable", "writable"]
    )
    assert len(parsed) == 1


@cu.register_parse_type("rev")
def _reverse_word(text):
    return "".join(reversed(text))


@pytest.mark.requires_netcdf
def test_parse_directory():
    df = cu.parse_directory(
        directories=[str(SAMPLES_DIR)],
        patterns=[
            "{activity}/{domain}/{institution}/{source}/{experiment}/{member:rev}/{frequency}/{?:_}.nc"
        ],
        homogenous_info={
            "mip_era": "CMIP6",
            "type": "simulation",
            "processing_level": "raw",
        },
        read_from_file=["variable", "date_start", "date_end", "version"],
        xr_open_kwargs={"engine": "h5netcdf"},
        cvs={
            "domain": {"example-region": "exreg"},
            "experiment": {"ssp126": {"driving_model": "driver"}},
            "attributes": {"version_id": "version"},
        },
        file_checks=["readable", "ncvalid"],
    )

    assert len(df) == 10
    assert (df["activity"] == "ScenarioMIP").all()
    assert (df["mip_era"] == "CMIP6").all()
    assert (df["domain"] == "exreg").all()  # CVS simple
    assert set(df["member"].unique()) == {"1f1p1i1r", "1f1p1i2r"}  # Custom parse type
    assert (
        df[df["frequency"] == "fx"]["variable"] == ("sftlf",)
    ).all()  # Read from file
    assert (
        df[df["experiment"] == "ssp126"]["driving_model"] == "driver"
    ).all()  # CVS complex
    assert (
        df[df["experiment"] != "ssp126"]["driving_model"].isnull().all()
    )  # CVS complex
    assert df.date_start.dtype == "<M8[ms]"
    assert df.date_end.dtype == "<M8[ms]"
    assert (
        df[df["frequency"] == "day"]["date_end"] == pd.Timestamp("2002-12-31")
    ).all()  # Read from file
    # Read from file + attrs cvs
    assert set(
        df[df["id"] == "CMIP6_ScenarioMIP_driver_NCC_NorESM2-MM_ssp126_1f1p1i1r_exreg"][
            "version"
        ]
    ) == {"v20191108", "v20200702"}


@pytest.mark.requires_netcdf
def test_parse_directory_readgroups():
    df = cu.parse_directory(
        directories=[str(SAMPLES_DIR)],
        patterns=[
            "{activity}/{domain}/{institution}/{source}/{experiment}/{member}/{frequency}/{?:_}.nc"
        ],
        read_from_file=[
            ["experiment", "frequency"],
            ["variable", "date_start", "date_end"],
        ],
        cvs={"variable": {"sftlf": None, "tas": "t2m"}},
    )
    assert len(df) == 10
    t2m = df.variable.apply(lambda v: "t2m" in v)
    assert (df[t2m]["date_end"] == pd.Timestamp("2002-12-31")).all()
    assert (df[~t2m].variable.apply(len) == 0).all()


@pytest.mark.requires_netcdf
def test_parse_directory_offcols():
    with pytest.raises(
        ValueError, match="Patterns include fields which are not recognized by xscen"
    ):
        cu.parse_directory(
            directories=[str(SAMPLES_DIR)],
            patterns=[
                "{activité}/{domain}/{institution}/{source}/{experiment}/{member}/{frequency}/{?:_}.nc"
            ],
        )
    df = cu.parse_directory(
        directories=[str(SAMPLES_DIR)],
        patterns=[
            "{activité}/{domain}/{institution}/{source}/{experiment}/{member}/{frequency}/{?:_}.nc"
        ],
        only_official_columns=False,
    )
    assert (df["activité"] == "ScenarioMIP").all()


@pytest.mark.requires_netcdf
def test_parse_directory_idcols():
    df = cu.parse_directory(
        directories=[str(SAMPLES_DIR)],
        patterns=[
            "{activity}/{domain}/{institution}/{source}/{experiment}/{member}/{frequency}/{?:_}.nc"
        ],
        only_official_columns=False,
        id_columns=["domain", "institution"],
    )
    assert (df["id"] == "example-region_NCC").all()


@pytest.mark.requires_netcdf
def test_parse_directory_skipdirs():
    df = cu.parse_directory(
        directories=[str(SAMPLES_DIR)],
        skip_dirs=[
            str(SAMPLES_DIR) + "/ScenarioMIP/example-region/NCC/NorESM2-MM/ssp126",
            str(SAMPLES_DIR) + "/ScenarioMIP/example-region/NCC/NorESM2-MM/ssp245/",
        ],
        patterns=[
            "{activity}/{domain}/{institution}/{source}/{experiment}/{member:rev}/{frequency}/{?:_}.nc"
        ],
        homogenous_info={
            "mip_era": "CMIP6",
            "type": "simulation",
            "processing_level": "raw",
        },
        read_from_file=["variable", "date_start", "date_end", "version"],
        xr_open_kwargs={"engine": "h5netcdf"},
        cvs={
            "domain": {"example-region": "exreg"},
            "attributes": {"version_id": "version"},
        },
        file_checks=["readable", "ncvalid"],
    )

    assert len(df) == 4
    assert (df["activity"] == "ScenarioMIP").all()
    assert (df["mip_era"] == "CMIP6").all()
    assert (df["domain"] == "exreg").all()  # CVS simple
    assert (
        df[df["frequency"] == "fx"]["variable"] == ("sftlf",)
    ).all()  # Read from file
    assert df.date_start.dtype == "<M8[ms]"
    assert df.date_end.dtype == "<M8[ms]"
    assert set(df.experiment.unique()) == {"ssp370", "ssp585"}


def test_parse_from_ds():
    # Real ds
    ds = xr.tutorial.open_dataset("air_temperature")
    info = cu.parse_from_ds(
        ds, names=["frequency", "source", "variable"], attrs_map={"platform": "source"}
    )
    assert info == {"frequency": "6hr", "source": "Model", "variable": ("air",)}


def test_parse_from_zarr(tmp_path):
    # Real ds
    ds = xr.tutorial.open_dataset("air_temperature")
    ds.to_zarr(tmp_path / "air.zarr")
    info = cu.parse_from_ds(
        tmp_path / "air.zarr",
        names=["frequency", "source", "variable"],
        attrs_map={"platform": "source"},
    )
    assert info == {"frequency": "6hr", "source": "Model", "variable": ("air",)}


def test_parse_from_nc(tmp_path):
    # Real ds
    ds = xr.tutorial.open_dataset("air_temperature")
    ds.to_netcdf(tmp_path / "air.nc", encoding={"air": {"dtype": "float32"}})
    info = cu.parse_from_ds(
        tmp_path / "air.nc",
        names=["frequency", "source", "variable"],
        attrs_map={"platform": "source"},
    )
    assert info == {"frequency": "6hr", "source": "Model", "variable": ("air",)}


def test_build_path(samplecat):
    df = cu.build_path(samplecat, root="/test", mip_era="CMIP5")
    assert (
        "/test/simulation/raw/CMIP5/ScenarioMIP/example-region/NCC/NorESM2-MM/ssp585/r1i1p1f1/fx/sftlf/"  # pragma: allowlist secret
        "sftlf_fx_CMIP5_ScenarioMIP_example-region_NCC_NorESM2-MM_ssp585_r1i1p1f1_fx.nc"
    ) in df.new_path.values


def test_pattern_from_schema(samplecat):
    df = cu.build_path(samplecat, mip_era="CMIP5")
    patts = cu.patterns_from_schema("original-sims-raw")
    for p in df.new_path.values:
        res = [cu._compile_pattern(patt).parse(p) for patt in patts]
        assert any(res)


@pytest.mark.parametrize("hasver", [True, False])
def test_build_path_ds(hasver):
    ds = xr.tutorial.open_dataset("air_temperature")
    ds = ds.assign(time=xr.cftime_range("0001-01-01", freq="6h", periods=ds.time.size))
    ds.attrs.update(source="source", institution="institution")
    if hasver:
        ds.attrs["version"] = "v1"
    new_path = cu.build_path(
        ds,
        schemas={
            "folders": [["source", "(version)"], "institution", ["variable", "xrfreq"]],
            "filename": ["source", "institution", "variable", "frequency", "DATES"],
        },
    )
    if hasver:
        assert new_path == Path(
            "source_v1/institution/air_6h/source_institution_air_6hr_0001-0002"
        )
    else:
        assert new_path == Path(
            "source/institution/air_6h/source_institution_air_6hr_0001-0002"
        )


def test_build_path_multivar(samplecat):
    info = samplecat.df.iloc[0].copy()
    info["variable"] = ("tas", "tasmin")
    with pytest.raises(
        ValueError,
        match="Selected schema original-sims-raw is meant to be used with single-variable datasets.",
    ):
        cu.build_path(info)
